####################################################################################################

__all__ = ["simulate_circuit"]

####################################################################################################

from InSpice import SpiceLibrary, Circuit, Simulator
# from InSpice import *
from InSpice.Unit import *

from InSpice.Doc.ExampleTools import find_libraries
libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

# 1/0

####################################################################################################

def simulate_circuit():
    circuit = Circuit('Capacitive Half-Wave Rectification (Pre Zener)')
    circuit.include(spice_library['1N4148'])
    circuit.include(spice_library['d1n5919brl'])
    ac_line = circuit.AcLine('input', 'L', circuit.gnd, rms_voltage=230@u_V, frequency=50@u_Hz)
    circuit.C('in', 'L', 1, 330@u_nF)
    circuit.R('emi', 'L', 1, 165@u_kΩ)
    circuit.R('in', 1, 2, 2*47@u_Ω)
    circuit.X('D1', '1N4148', 2, 'out')
    circuit.C('2', 'out', 3, 250@u_uF)
    circuit.R('2', 3, circuit.gnd, 1@u_kΩ)
    circuit.X('D2', '1N4148', 3, 2)
    circuit.X('Dz', 'd1n5919brl', circuit.gnd, 'out')
    circuit.C('', 'out', circuit.gnd, 250@u_uF)
    circuit.R('load', 'out', circuit.gnd, 1@u_kΩ)

    simulator = Simulator.factory()
    simulation = simulator.simulation(circuit, temperature=25, nominal_temperature=25)
    analysis = simulation.transient(step_time=ac_line.period/200, end_time=ac_line.period*50)

    return analysis



####################################################################################################

import numpy as np

import matplotlib.pyplot as plt

####################################################################################################

import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################
from InSpice.Doc.ExampleTools import find_libraries
from InSpice.Plot.BodeDiagram import bode_diagram
from InSpice import SpiceLibrary, Circuit, Simulator, plot
from InSpice.Unit import *

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

#f# literal_include('OperationalAmplifier.py')

####################################################################################################

circuit = Circuit('Operational Amplifier')
circuit.include(spice_library['LMH6551'])
gain = 1/2
rh_val = 11@u_kΩ
vdd_voltage = 5 # max voltage for the LMH6551 is 13.5V

circuit.SinusoidalVoltageSource('input', 'in', circuit.gnd, offset = 0, amplitude=1@u_V, frequency=100@u_kHz)
circuit.V('Vcm', 'Vcm', circuit.gnd, dc_value=(vdd_voltage/2)@u_V)
circuit.V('V+', 'V+', circuit.gnd, dc_value=vdd_voltage@u_V)
# place a 0.1uF capacitor from V+ to ground
circuit.C('V+', 'V+', circuit.gnd, 0.1@u_uF)
# place a 0.1uF capacitor from Vcm to ground
circuit.C('Vcm', 'Vcm', circuit.gnd, 0.1@u_uF)

circuit.X('op', 'LMH6551', 'in+', 'in-', 'V+', circuit.gnd, 'Vcm', 'out+', 'out-')
circuit.R('inph', 'in',          'in+', rh_val)
circuit.R('innh', circuit.gnd,   'in-', rh_val)

circuit.R('innl', 'in+', 'out-', gain*rh_val)
circuit.R('inpl', 'in-', 'out+', gain*rh_val)

# put a 56 omhm in series with each output of th opamp
circuit.R('out+', 'out+', 'outp', 56@u_Ω)
circuit.R('out-', 'out-', 'outn', 56@u_Ω)
# place 39pf from outp and outn to ground
circuit.C('outp', 'outp', circuit.gnd, 39@u_pF)
circuit.C('outn', 'outn', circuit.gnd, 39@u_pF)

circuit.R('load', 'outp', 'outn', 470@u_Ω)


simulator = Simulator.factory()
simulation = simulator.simulation(circuit, temperature=25, nominal_temperature=25)
analysis = simulation.transient(step_time=1e-6, end_time=20e-6, start_time=0@u_ms, max_time=1@u_ns, use_initial_condition=True)

figure, (ax1, ax2) = plt.subplots(2, figsize=(20, 10))

plt.title("output Operational Amplifier")
plt.xlabel("Time (s)")
plt.ylabel("Voltage (V)")
ax1.plot(analysis.time, analysis['in'], label='input voltage')
ax1.plot(analysis.time, analysis['outp']-analysis['outn'], label='differential output')

# add legend
ax1.legend()

# ax2.plot(analysis.time, analysis['in+'], label='opamp in+')
# ax2.plot(analysis.time, analysis['in-'], label='opamp in-')
ax2.plot(analysis.time, analysis['outp'], label='opamp out+')
ax2.plot(analysis.time, analysis['outn'], label='opamp out-')
# add legend
ax2.legend()

plt.tight_layout()
plt.show()

#f# save_figure('figure', 'operational-amplifier.png')



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = ['expand_path', 'find', 'walk']

####################################################################################################

from typing import Iterator

import os
from pathlib import Path

####################################################################################################

def find(file_name: str, directories: list[str]) -> Path:
    # Fixme: bytes ???
    #  on Linux path are bytes, thus some files can be invalid utf8...
    # if isinstance(directories, bytes):
    #     directories = (directories,)
    for directory in directories:
        directory = Path(directory)
        # for directory_path, _, file_names in os.walk(directory):
        for directory_path, _, file_names in directory.walk():
            directory_path = Path(directory_path)
            if file_name in file_names:
                return directory_path.joinpath(file_name)
    raise NameError(f"File {file_name} not found in directories {directories}")

####################################################################################################

def expand_path(path: Path | str) -> Path:
    # Substrings of the form $name or ${name} are replaced by the value of environment variable name.
    # On Unix and Windows, return the argument with an initial component of ~ or ~user
    # replaced by that user’s home directory.
    _ = os.path.expandvars(path)
    return Path(_).expanduser().absolute()

####################################################################################################

def walk(path: Path | str, followlinks: bool = False) -> Iterator[Path]:
    for root, _, files in Path(path).walk(follow_symlinks=followlinks):
        for filename in files:
            yield Path(root).joinpath(filename)



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = ["Hdf5File"]

####################################################################################################

# File structure
#   using h5ls -r simulation.hdf5
#
# /                        Group
# /version                 Dataset {SCALAR}
# /simulation              Group
# /abscissas               Group
# /abscissas/time          Dataset {10038}
# /branches                Group
# /branches/e.xdz.ev1      Dataset {10038}
# /branches/...
# /elements                Group
# /internal_parameters     Group
# /nodes                   Group
# /nodes/1                 Dataset {10038}
# /nodes/...

# h5dump -g "/simulation" simulation.hdf5
# h5dump -d "/abscissas/time" simulation.hdf5
# h5dump -d "/nodes/1" simulation.hdf5

####################################################################################################

import pickle

import numpy as np

# https://docs.h5py.org/en/stable/index.html
import h5py

####################################################################################################

class Hdf5File:

    _VERSION = 1

    ##############################################

    @classmethod
    def save(cls, analysis, path):
        self = cls(path, 'w')
        self._save(analysis)

    ##############################################

    def __init__(self, path, mode=None):
        self._file = h5py.File(path, mode)

    ##############################################

    def _save(self, analysis):
        self._file["version"] = self._VERSION

        simulation = analysis.simulation
        group = self._file.create_group("simulation")
        attrs = group.attrs
        attrs["simulator"] = simulation.simulator.SIMULATOR
        attrs["simulator_version"] = simulation.simulator_version
        attrs["simulation_date"] = str(simulation.simulation_date)
        attrs["simulation_duration"] = str(simulation.simulation_duration)
        attrs["spice_circuit"] = str(simulation.circuit)
        attrs["spice_simulation"] = str(simulation) # Fixme: remove desk
        # https://docs.h5py.org/en/stable/strings.html#how-to-store-raw-binary-data
        attrs["pickled_simulation"] = np.void(pickle.dumps(simulation))

        self._abscissa = self._file.create_group("abscissas")
        for list_name in ("nodes", "branches", "elements", "internal_parameters"):
            group = self._file.create_group(list_name)
            for waveform in getattr(analysis, list_name).values():
                self._add_waveform(group, waveform)

    ##############################################

    def _add_waveform_common(self, group, waveform):
        dset = group.create_dataset(
            waveform.name,
            data=waveform,
            compression="gzip", compression_opts=9
            # compression="lzf
        )
        attrs = dset.attrs
        attrs["unit"] = str(waveform.unit)
        return dset

    ##############################################

    def _add_abscissa(self, abscissa):
        # Fixme: is time always unique ???
        # See https://docs.h5py.org/en/stable/high/group.html#hard-links
        if abscissa.name not in self._abscissa:
            # self._abscissa[abscissa.name] = abscissa
            self._add_waveform_common(self._abscissa, abscissa)
        return self._abscissa[abscissa.name]

    ##############################################

    def _add_waveform(self, group, waveform):
        dset = self._add_waveform_common(group, waveform)
        attrs = dset.attrs
        if waveform.title:
            attrs["title"] = waveform.title
        attrs["abscissa"] = self._add_abscissa(waveform.abscissa).ref



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = ['Builder', 'ToPython']

####################################################################################################

import logging
import os

from InSpice.Spice.BasicElement import SubCircuitElement
from InSpice.Spice.Element import ElementParameterMetaClass
from InSpice.Spice.ElementParameter import FlagParameter
from InSpice.Spice.Netlist import Circuit, SubCircuit, Node
from .HighLevelParser import *

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class Translator:

    ##############################################

    def handle(self, obj):
        cls = obj.__class__.__name__
        handler = getattr(self, f'handle_{cls}')
        return handler(obj)

####################################################################################################

class Builder(Translator):

    _logger = _module_logger.getChild('Builder')

    ##############################################

    def __init__(self) -> None:
        self._circuit = None
        self._ground = None

    ##############################################

    @property
    def circuit(self) -> Circuit:
        return self._circuit

    ##############################################

    def translate(self, spice_code:SpiceSource, ground: int=Node.SPICE_GROUND_NUMBER) -> Circuit:
        """Build a :class:`Circuit` instance.

        Use the *ground* parameter to specify the node which must be translated to 0 (SPICE ground node).

        """
        self._circuit = Circuit(spice_code.title)
        self._ground = ground
        
        for obj in spice_code.obj_lines:
            self.handle(obj)

        return self._circuit

    ##############################################

    def handle_Ac(self, obj: Ac) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Control(self, obj: Control) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Csparam(self, obj: Csparam) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Dc(self, obj: Dc) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Distorsion(self, obj: Distorsion) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Element(self, obj: Element) -> None:
        factory = getattr(self._circuit, obj.factory.ALIAS)
        nodes = obj.translate_ground_node(self._ground)
        parameters = [str(_) for _ in obj._parameters]
        kwargs = obj.dict_parameters
        if obj.letter != 'X':
            # args = nodes + parameters
            args = nodes
            kwargs['raw_spice'] = ' '.join(parameters)
        else:    # != Spice
            args = parameters + nodes
        self._logger.debug(str((obj.letter, obj.name, args, kwargs)))
        self._logger.debug(str((nodes, parameters)))
        # factory(obj.name, *args, **kwargs)
        factory(obj.name, *args, **kwargs)

    ##############################################

    def handle_Else(self, obj: Else) -> None:
        raise NotImplementedError

    ##############################################

    def handle_ElseIf(self, obj: ElseIf) -> None:
        raise NotImplementedError

    ##############################################

    def handle_EndControl(self, obj: EndControl) -> None:
        raise NotImplementedError

    ##############################################

    def handle_End(self, obj: End) -> None:
        pass

    ##############################################

    def handle_EndIf(self, obj: EndIf) -> None:
        raise NotImplementedError

    ##############################################

    def handle_EndSubcircuit(self, obj: EndSubcircuit) -> None:
        raise NotImplementedError
        # pass

    ##############################################

    def handle_Fourier(self, obj: Fourier) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Function(self, obj: Function) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Global(self, obj: Global) -> None:
        raise NotImplementedError

    ##############################################

    def handle_If(self, obj: If) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Include(self, obj: Include) -> None:
        self._circuit.include(obj.path)

    ##############################################

    def handle_InitialCondition(self, obj: InitialCondition) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Library(self, obj: Library) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Measure(self, obj: Measure) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Model(self, obj: Model) -> None:
        self._circuit.model(obj.name, obj.type, **obj.parameters)

    ##############################################

    def handle_NodeSet(self, obj: NodeSet) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Noise(self, obj: Noise) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Op(self, obj: Op) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Options(self, obj: Options) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Param(self, obj: Param) -> None:
        raise NotImplementedError

    ##############################################

    def handle_PeriodicSteadyState(self, obj: PeriodicSteadyState) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Plot(self, obj: Plot) -> None:
        raise NotImplementedError

    ##############################################

    def handle_PoleZero(self, obj: PoleZero) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Print(self, obj: Print) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Probe(self, obj: Probe) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Save(self, obj: Save) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Sensitivity(self, obj: Sensitivity) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Subcircuit(self, obj: Subcircuit, ground=Node.SPICE_GROUND_NUMBER) -> None:
        subcircuit = SubCircuit(obj._name, *obj._nodes)
        # Add all elements from obj._items to the subcircuit
        for item in obj._items:
            # Create a temporary Builder to handle each item in context of the subcircuit
            temp_builder = Builder()
            temp_builder._circuit = subcircuit
            temp_builder._ground = ground
            temp_builder.handle(item)
        # Original SpiceParser._build_circuit(subcircuit, obj._statements, ground)
        self._circuit.subcircuit(subcircuit)        
        return subcircuit

    ##############################################

    def handle_Temperature(self, obj: Temperature) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Title(self, obj: Title) -> None:
        self._circuit.title = str(obj)

    ##############################################

    def handle_Transient(self, obj: Transient) -> None:
        raise NotImplementedError

    ##############################################

    def handle_Width(self, obj: Width) -> None:
        raise NotImplementedError

####################################################################################################

class ToPython(Translator):

    _logger = _module_logger.getChild('ToPython')

    ##############################################

    @classmethod
    def value_to_python(cls, x) -> str:
        if x:
            match x:
                case int() | float():
                    return str(x)
                case _:
                    return f"'{x}'"
        else:
            return ''

    ##############################################

    @classmethod
    def values_to_python(cls, values: list) -> list[str]:
        return [cls.value_to_python(x) for x in values]

    ##############################################

    @classmethod
    def format_kwargs(cls, kwargs: dict) -> list[str]:
        return [f'{key}={cls.value_to_python(value)}' for key, value in kwargs.items()]

    ##############################################

    @classmethod
    def join_args(cls, args: list) -> str:
        return ', '.join(args)

    ##############################################

    def __init__(self) -> None:
        self._ground = None
        self._circuit_name = 'circuit'

    ##############################################

    def translate(self, spice_code:SpiceSource, ground=Node.SPICE_GROUND_NUMBER) -> str:
        self._ground = str(ground)
        source_code = ''

        # Fixme: if self.circuit:
        title = spice_code.title or ''
        source_code += f"circuit = Circuit('{title}')" + os.linesep
        for obj in spice_code.obj_lines:
            line = self.handle(obj)
            if line is not None:
                source_code += line + os.linesep

        return source_code

    ##############################################

    def handle_Ac(self, obj: Ac) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Control(self, obj: Control) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Csparam(self, obj: Csparam) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Dc(self, obj: Dc) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Distorsion(self, obj: Distorsion) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Element(self, obj: Element) -> str:
        args = [obj.name]
        nodes = obj.translate_ground_node(self._ground)
        parameters = [str(_) for _ in obj._parameters]
        if obj.letter != 'X':
            args += nodes + parameters
        else:    # != Spice
            args += parameters + nodes
        args = self.values_to_python(args)
        kwargs = self.format_kwargs(obj.dict_parameters)
        parameters = self.join_args(args + kwargs)
        return f'{self._circuit_name}.{obj.letter}({parameters})'

    ##############################################

    def handle_Else(self, obj: Else) -> str:
        raise NotImplementedError

    ##############################################

    def handle_ElseIf(self, obj: ElseIf) -> str:
        raise NotImplementedError

    ##############################################

    def handle_EndControl(self, obj: EndControl) -> str:
        raise NotImplementedError

    ##############################################

    def handle_End(self, obj: End) -> str:
        return None

    ##############################################

    def handle_EndIf(self, obj: EndIf) -> str:
        raise NotImplementedError

    ##############################################

    def handle_EndSubcircuit(self, obj: EndSubcircuit) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Fourier(self, obj: Fourier) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Function(self, obj: Function) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Global(self, obj: Global) -> str:
        raise NotImplementedError

    ##############################################

    def handle_If(self, obj: If) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Include(self, obj: Include) -> str:
        return f'{self._circuit_name}.include({obj})'

    ##############################################

    def handle_InitialCondition(self, obj: InitialCondition) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Library(self, obj: Library) -> str:
        return f'{self._circuit_name}.lib({obj})'

    ##############################################

    def handle_Measure(self, obj: Measure) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Model(self, obj: Model) -> str:
        args = self.values_to_python((obj.name, obj.type))
        kwargs = self.format_kwargs(obj.py_parameters)
        parameters = self.join_args(args + kwargs)
        return f'{self._circuit_name}.model({parameters})'

    ##############################################

    def handle_NodeSet(self, obj: NodeSet) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Noise(self, obj: Noise) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Op(self, obj: Op) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Options(self, obj: Options) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Param(self, obj: Param) -> str:
        raise NotImplementedError

    ##############################################

    def handle_PeriodicSteadyState(self, obj: PeriodicSteadyState) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Plot(self, obj: Plot) -> str:
        raise NotImplementedError

    ##############################################

    def handle_PoleZero(self, obj: PoleZero) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Print(self, obj: Print) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Probe(self, obj: Probe) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Save(self, obj: Save) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Sensitivity(self, obj: Sensitivity) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Subcircuit(self, obj: Subcircuit, ground=Node.SPICE_GROUND_NUMBER) -> str:
        raise NotImplementedError
        # subcircuit_name = 'subcircuit_' + obj.name
        # args = self.values_to_python([subcircuit_name] + obj.nodes)
        # source_code = ''
        # source_code += '{} = SubCircuit({})'.format(subcircuit_name, obj.join_args(args)) + os.linesep
        # source_code += SpiceParser.netlist_to_python(subcircuit_name, obj, ground)
        # return source_code

    ##############################################

    def handle_Temperature(self, obj: Temperature) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Title(self, obj: Title) -> str:
        return f"{self._circuit_name} = (str(obj))"

    ##############################################

    def handle_Transient(self, obj: Transient) -> str:
        raise NotImplementedError

    ##############################################

    def handle_Width(self, obj: Width) -> str:
        raise NotImplementedError



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = ['elements']

####################################################################################################

import logging

####################################################################################################

from InSpice.Spice.Element import ElementParameterMetaClass
from InSpice.Spice.ElementParameter import FlagParameter

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class ElementData:

    """This class represents a device prefix letter."""

    ##############################################

    def __init__(self, letter: str, classes) -> None:

        # Fixme: _ & property

        self.letter = letter
        self.classes = classes

        number_of_positionals_min = 1000
        number_of_positionals_max = 0
        has_optionals = False
        for element_class in classes:
            number_of_positionals = element_class.number_of_positional_parameters
            number_of_positionals_min = min(number_of_positionals_min, number_of_positionals)
            number_of_positionals_max = max(number_of_positionals_max, number_of_positionals)
            has_optionals = max(has_optionals, bool(element_class.optional_parameters))

        self.number_of_positionals_min = number_of_positionals_min
        self.number_of_positionals_max = number_of_positionals_max
        self.has_optionals = has_optionals

        self.multi_devices = len(classes) > 1
        self.has_variable_number_of_pins = letter in ('Q', 'X') # NPinElement, Q has 3 to 4 pins
        if self.has_variable_number_of_pins:
            self.number_of_pins = None
        else:
            # Q and X are single
            self.number_of_pins = classes[0].number_of_pins

        self.has_flag = False
        for element_class in classes:
            for parameter in element_class.optional_parameters.values():
                if isinstance(parameter, FlagParameter):
                    self.has_flag = True

    ##############################################

    def __len__(self) -> int:
        return len(self.classes)

    ##############################################

    def __iter__(self):
        return iter(self.classes)

    ##############################################

    @property
    def single(self):
        if not self.multi_devices:
            return self.classes[0]
        else:
            raise NameError()

####################################################################################################

elements = {}
def _init() -> None:
    for letter, classes in ElementParameterMetaClass._classes.items():
        element_data = ElementData(letter, classes)
        elements[letter] = element_data
        elements[letter.lower()] = element_data
_init()

####################################################################################################

if __name__ == '__main__':
    for element_data in sorted(elements.values(), key=lambda x: len(x)):
        print(element_data.letter,
              len(element_data),
              element_data.number_of_positionals_min, element_data.number_of_positionals_max,
              element_data.has_optionals)

# Single:
# B 0 True
# D 1 True
# F 2 False
# G 1 False
# H 2 False
# I 1 False
# J 1 True
# K 3 False
# M 1 True
# S 2 False
# V 1 False
# W 3 False
# Z 1 True

# Two:
# E 0 1 False
# L 1 2 True

# Three:
# C 1 2 True
# R 1 2 True

# NPinElement:
# Q 1 1 True
# X 1 1 False



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

from pathlib import Path
from typing import Iterable, Iterator

import logging
import os
import pickle
import re

from InSpice.Spice.Parser import Subcircuit, Model
from InSpice.Tools import PathTools
from .SpiceInclude import SpiceInclude, is_yaml

####################################################################################################

NEWLINE = os.linesep

_module_logger = logging.getLogger(__name__)

####################################################################################################

class SpiceLibrary:

    """This class implements a Spice sub-circuits and models library.

    A library is a directory which is recursively scanned for '.lib' file and parsed for sub-circuit
    and models definitions.

    Example of usage::

        spice_library = SpiceLibrary('/some/path/')

    If the directory hierarchy contains a file that define a 1N4148 sub-circuit then we can retrieve
    the file path using::

        spice_library['1N4148']

    """

    _logger = _module_logger.getChild('Library')

    EXTENSIONS = (
        '.spice',
        '.lib',
        '.cir',
        '.mod',
        '.lib@xyce',
        '.mod@xyce',
    )

    ##############################################

    def __init__(self, root_path: str | Path, scan: bool = False, recurse: bool = False, section: bool = False) -> None:
        # recurse will be removed in the future maybe. it's here because skidl uses it
        # if recurse:
        #     scan = recurse
        self._path = PathTools.expand_path(root_path)
        if not self._path.exists():
            self._path.mkdir(parents=True)
            self._logger.info(f"Created {self._path}")
        # elif self._path.is_file():
        #     self._path = self._path.parent
        self._subcircuits = {}
        self._models = {}
        self._recurse = recurse
        self._section = section
        if not scan:
            if self.has_db_path:
                self.load()
                '''Check if the library has the our path in the subcircuits.'''
                paths = {Path(sub_path) for sub_path in self._subcircuits.values()}
                if self._path not in paths:
                    scan = True 
            else:
                self._logger.info("Initialize library...")
                scan = True
        if scan:
            self.scan()
            self.save()

    ##############################################

    @property
    def db_path(self) -> Path:
        if self._path.is_file():
            # If the db path is for a file, use the file's parent directory
            return self._path.parent.joinpath('db.pickle')
        return self._path.joinpath('db.pickle')

    @property
    def has_db_path(self) -> bool:
        return self.db_path.exists()

    ##############################################

    def __bool__(self) -> bool:
        return bool(self._subcircuits or self._models)

    ##############################################

    def __getstate__(self):
        # state = self.__dict__.copy()
        state = {
            'subcircuits': self._subcircuits,
            'models': self._models,
        }
        return state

    ##############################################

    def __setstate__(self, state):
        # self.__dict__.update(state)
        self._subcircuits = state['subcircuits']
        self._models = state['models']

    ##############################################

    def save(self) -> None:
        with open(self.db_path, 'wb') as fh:
            _ = self.__getstate__()
            pickle.dump(_, fh)

    def load(self) -> None:
        self._logger.info(f"Load {self.db_path}")
        with open(self.db_path, 'rb') as fh:
            _ = pickle.load(fh)
            self.__setstate__(_)

    ##############################################

    def _category_path(self, category: str) -> Path:
        category = category.split('/')
        return self._path.joinpath(*category)

    ##############################################

    def add_category(self, category: str) -> None:
        path = self._category_path(category)
        if not path.exists():
            path.mkdir(parents=True)
            self._logger.info(f"Created {path}")
        else:
            self._logger.info(f"category '{category}' already exists")

    ##############################################

    def _list_categories(self, path: Path | str, level: int = 0) -> str:
        text = ''
        indent = ' '*4*level
        for entry in sorted(os.scandir(path), key=lambda entry: entry.name):
            if entry.is_dir():
                text += f'{indent}{entry.name}' + NEWLINE
                text += self._list_categories(entry.path, level+1)
        return text

    def list_categories(self) -> str:
        return self._list_categories(self._path)

    ##############################################

    def scan(self) -> None:
        
        self._logger.info(f"Scan {self._path}...")
        
        # Handle the case where self._path is a file, not a directory
        if self._path.is_file():
            # Check if the file has a valid extension
            _ = self._path.suffix.lower()
            if _ in self.EXTENSIONS:
                try:
                    self._handle_library(self._path)
                except Exception as e:
                    self._logger.warning(f"Failed to parse {self._path}: {e}")
            return
                
        # Handle the case where self._path is a directory
        for path in PathTools.walk(self._path):
            _ = path.suffix.lower()
            if _ in self.EXTENSIONS:
                try:
                    self._handle_library(path)
                except Exception as e:
                    self._logger.warning(f"Failed to parse {path}: {e}")

    ##############################################

    def _handle_library(self, path: Path) -> None:
        spice_include = SpiceInclude(path, recurse=self._recurse, section=self._section)
        # Fixme: check overwrite
        self._models.update({_.name: str(_.path) for _ in spice_include.models})
        self._subcircuits.update({_.name: str(_.path) for _ in spice_include.subcircuits})

    ##############################################

    def delete_yaml(self) -> None:
        for path in PathTools.walk(self._path):
            if is_yaml(path):
                self._logger.info(f"{NEWLINE}Delete {path}")
                path.unlink()

    ##############################################

    def __getitem__(self, name: str) -> Subcircuit | Model:
        if not self:
            self._logger.warning("Empty library")
        
        # First, check if the requested item exists directly
        path = None
        if name in self._subcircuits:
            path = self._subcircuits[name]
        elif name in self._models:
            path = self._models[name]
        else:
            # Item not found directly - warn and raise KeyError
            available = list(self._subcircuits.keys()) + list(self._models.keys())
            available_str = ", ".join(available[:10])
            if len(available) > 10:
                available_str += f", ... ({len(available)-10} more)"
            self._logger.warning(f"Library item '{name}' not found in {self._path}. Available: {available_str}")
            raise KeyError(name)
        
        # Create SpiceInclude with recursion enabled if requested
        spice_include = SpiceInclude(path, recurse=self._recurse)
        
        try:
            # Try to get the item directly from this SpiceInclude
            return spice_include[name]
        except KeyError:
            # Item exists in index but not in the file - search in parent directory
            self._logger.info(f"Item '{name}' referenced in {path} but not found there. Searching in sibling files...")
            
            # Try to find the component in sibling library files
            original_path = Path(path)
            parent_dir = original_path.parent
            
            # Try looking for the item in other library files in the same directory
            for lib_ext in self.EXTENSIONS:
                for sibling_file in parent_dir.glob(f"*{lib_ext}"):
                    if sibling_file == original_path:
                        continue  # Skip the original file
                        
                    try:
                        self._logger.info(f"Checking {sibling_file} for {name}")
                        sibling_include = SpiceInclude(sibling_file, recurse=self._recurse)
                        # Check if this file contains our item
                        for subckt in sibling_include.subcircuits:
                            if subckt.name == name:
                                return subckt
                        for model in sibling_include.models:
                            if model.name == name:
                                return model
                    except Exception as e:
                        self._logger.warning(f"Error checking {sibling_file}: {e}")
            
            # If we still can't find it, try one last method - force reparse everything
            try:
                self._logger.info(f"Attempting one last search with forced reparse for {name}")
                reparse_include = SpiceInclude(path, rewrite_yaml=True, recurse=True)
                return reparse_include[name]
            except KeyError:
                # Detailed error message when all recovery attempts fail
                raise KeyError(f"'{name}' referenced in library index but not found in any source file. Check your include hierarchy.")

    ##############################################

    @property
    def subcircuits(self) -> Iterator[Subcircuit]:
        """ Dictionary of sub-circuits """
        return iter(self._subcircuits)

    @property
    def models(self) -> Iterator[Model]:
        """ Dictionary of models """
        return iter(self._models)

    # ##############################################

    # def iter_on_subcircuits(self):
    #     return self._subcircuits.itervalues()

    # ##############################################

    # def iter_on_models(self):
    #     return self._models.itervalues()

    # ##############################################

    def search(self, regexp: str) -> Iterable[tuple[str, SpiceInclude]]:
        """ Return dict of all models/subcircuits with names matching regex. """
        regexp = re.compile(regexp)
        models_subcircuits = {**self._models, **self._subcircuits}
        if not models_subcircuits:
            self._logger.warning("Empty library")
        for name, _ in models_subcircuits.items():
            if regexp.search(name):
                yield name, SpiceInclude(_)



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module tries to implements a LALR parser for Spice.

**Notes**

* Ngspice only use a LALR parser to parse mathematical expression.
* Ngspice syntax is sometimes contextual.
* We can implement a tokenizer.
* But the Ngspice's rules to group the tokens is unknown, i.e. how to handle :code:`() {} [] '' =`
  and to make the AST.
* Moreover it is more difficult to implement by hand such algorithm.  Some experimental codes was
  removed in favour of this LALR parser.

**Known Syntax Issues**

* :code:`in_offset=[0.1 -0.2]` is interpreted as a subtraction due to the space token separator.  We
  could fix that by adding the unary minus in the float regexp, but it would break the parsing of
  mathematical expressions.  A workaround is to use a brace expression, :code:`{-0.2}`.  Can we fix
  the AST afterwards ???  For example, remove recursively the unary minus from the AST.

"""

__all__ = ['SpiceParser']

####################################################################################################
# Fixme:
#
#  Valid syntax ???
#  print res .endc
####################################################################################################

####################################################################################################

import logging
from typing import Optional

import ply.lex as lex
import ply.yacc as yacc

from .Ast import *

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class SpiceParser:

    _logger = _module_logger.getChild('SpiceParser')

    ##############################################

        # Declare an INCLUSIVE state called PARAM
    states = (
        ('NODES', 'inclusive'),
    )

    ###############################################

    # When building the master regular expression, rules are added in the following order:
    #   - All tokens defined by functions are added in the same order as they appear in the lexer file.
    #   - Tokens defined by strings are added next by sorting them in order of decreasing regular
    #     expression length (longer expressions are added first).

    tokens = [
        # 'END_OF_LINE_COMMENT',

        # in this order
        'ID',
        'NUMBER',

        # sorted by length
        'MINUS',
        'NOT',
        'POWER',
        'MULTIPLY',
        'DIVIDE',
        'MODULO',
        'INT_DIVIDE',
        'ADD',
        'EQUAL',
        'NOT_EQUAL',
        'LESS_EQUAL',
        'GREATER_EQUAL',
        'LESS',
        'GREATER',
        'AND',
        'OR',
        'QUESTION', 'COLON',

        'LEFT_PARENTHESIS', 'RIGHT_PARENTHESIS',
        'COMMA',

        'LEFT_BRACKET', 'RIGHT_BRACKET',

        'LEFT_BRACE', 'RIGHT_BRACE',
        'QUOTE',
        'SET',

        'BRANCH',
        'AT',

        'TILDE',

        'STRING',

        'DOT_COMMAND',
    ]

    ##############################################

    def t_error(self, token):
        self._logger.error(f"Illegal character '{token.value[0]}' at line {token.lexer.lineno} and position {token.lexer.lexpos}")
        # token.lexer.skip(1)
        raise NameError('Lexer error')


    def t_DOT_COMMAND(self, t):
        r'\.(?i:[a-z]+)'
        # Switch to the PARAM state on seeing a dot command
        t.lexer.begin('NODES')
        return t
    
    def t_NODES_SET(self, t):
        r'='
        # Switch back to the DEFAULT state on seeing a =
        t.lexer.begin('INITIAL')
        return t
    ##############################################

    t_ignore = ' \t'

    ##############################################

    # t_END_OF_LINE_COMMENT = r';|\$'

    # Fixme: could use
    # literals = "+-*/"

    t_MINUS = r'-'
    t_NOT = r'!'
    t_POWER = r'(\*\*)|\^'
    t_MULTIPLY = r'\*'
    t_DIVIDE = r'/'
    t_MODULO = r'%'
    t_INT_DIVIDE = r'\\'
    t_ADD = r'\+'
    t_EQUAL = r'=='
    t_NOT_EQUAL = r'!='
    t_LESS_EQUAL = r'<='
    t_GREATER_EQUAL = r'>='
    t_LESS = r'<'
    t_GREATER = r'>'
    t_AND = r'&&'
    t_OR = r'\|\|'
    t_QUESTION = r'\?'
    t_COLON = r':'

    t_LEFT_PARENTHESIS = r'\('
    t_RIGHT_PARENTHESIS = r'\)'
    t_COMMA = r','

    t_LEFT_BRACKET = r'\['
    t_RIGHT_BRACKET = r'\]'

    t_LEFT_BRACE = r'\{'
    t_RIGHT_BRACE = r'\}'

    t_QUOTE = r'[\'"’]'
    t_SET = r'='

    t_BRANCH = r'\#branch'
    t_AT = r'@'

    t_TILDE = r'~'

    # t_STRING = r'"((\\")|[^"])*"'

    # t_DOT_COMMAND = r'\.(?i:[a-z]+)'

    # Note:
    #      If ID > NUMBER, then it breaks float and yield A - B
    #      If NUMBER < ID, then 2N2222A is split in two NUMBER tokens: Number 2 n, Number 2222 None a
    #   Solution: use [a-z0-9]* for EXTRA_UNIT

    
    def t_NUMBER(self, t):
        # Fixme: CONTEXTUAL SYNTAX !!! in_offset=[0.1 -0.2]
        r'''
        (?i:
            (?P<NUMBER_PART>
                # [-+]?
                (?:
                    (?: \d* \. \d+ )   # .1 .12 ... 9.1 98.1 ...
                    |
                    (?: \d+ \.? )      # 1. 12. ... 1 12 ...
                )
                (?: e [+-]? \d+ ) ?
            )
            (?P<UNIT_PART>
                (meg) | (mil) | [tgkmunpf]
            ) ?
            (?P<EXTRA_UNIT>
                [a-z0-9]*
            )
        )
        '''
        match = t.lexer.lexmatch.groupdict()
        value = match['NUMBER_PART']
        unit = match['UNIT_PART']
        extra_unit = match['EXTRA_UNIT']
        # Check case: 2 n 2222a
        if unit:
            for _ in extra_unit:
                if _.isdigit():
                    t.type = 'ID'
                    t.value = t.lexer.lexmatch[0]
                    return t
        try:
            value = int(value)
        except ValueError:
            value = float(value)
        if isinstance(value, int) and not unit and not extra_unit:
            t.value = Integer(value)
        else:
            t.value = Number(value, unit, extra_unit)
        return t
    def t_ID(self, t):
        # Fixme:
        r'''
        (?i:
            [a-z_0-9]+
            (\.[a-z_0-9.]+) ?
        )'''
        # t.value = Id(t.value)
        return t
    

    def t_NODES_ID(self, t):
        # Fixme:
        r'''
        (?i:
            (?:[+\-\%](?=[a-z_]))?
            [a-z0-9_][-a-z0-9_]*
            (?:\.[a-z0-9_][-a-z0-9_]*)?
            (?:(?<=[a-z0-9])[+\-](?![a-z0-9.-]))?
            :?
        )'''
        # t.value = Id(t.value)
        return t


    ##############################################
    #
    # Grammar
    #

    # from lowest
    precedence = (
        ('left', 'QUESTION', 'COLON'),   # Fixme: left ?
        ('left', 'OR'),
        ('left', 'AND'),
        ('nonassoc', 'EQUAL', 'NOT_EQUAL', 'LESS_EQUAL', 'GREATER_EQUAL', 'LESS', 'GREATER'),
        ('left', 'ADD', 'MINUS'),
        ('left', 'MULTIPLY', 'DIVIDE', 'MODULO', 'INT_DIVIDE'),
        ('left', 'POWER'),
        ('right', 'UMINUS', 'NOT'),
    )

    # Normally, the first rule found in a yacc specification defines the starting grammar rule
    # (top level rule). To change this, simply supply a start specifier in your file.
    # start = ''

    def _command(self, p, cls):
        if len(p) == 3:
            p[0] = cls(p[1], p[2])
        else:
            p[0] = cls(p[1])
        return p[0]

    def p_command(self, p):
        '''command : ID expression_list_space
                   | ID
        '''
        return self._command(p, Command)

   
    # Fixme: could be merged with command
    def p_dot_command(self, p):
        '''command : DOT_COMMAND expression_list_space
                   | DOT_COMMAND
        '''
        return self._command(p, DotCommand)

    def p_branch(self, p):
        '''expression : ID BRANCH
        '''
        p[0] = Branch(p[1])

    def p_tilde(self, p):
        # Fixme: node, number is a node integer
        '''expression : TILDE ID
                      | TILDE NUMBER
        '''
        p[0] = InvertedInput(p[2])

    def p_inner_parameter(self, p):
        '''expression : AT ID LEFT_BRACKET ID RIGHT_BRACKET
        '''
        p[0] = InnerParameter(p[2], p[4])

    def p_number(self, p):
        '''expression : NUMBER
        '''
        p[0] = p[1]

    def p_id(self, p):
        '''expression : ID
        '''
        p[0] = Id(p[1])
    
    def p_uminus(self, p):
        '''expression : MINUS expression %prec UMINUS'''
        # %prec UMINUS overrides the default rule precedence-setting it to that of UMINUS in the precedence specifier.
        _ = p[2]
        if isinstance(_, (Integer, Number)):
            p[0] = _.negate()
        else:
            p[0] = Negation(_)

    def p_unnary_operation(self, p):
        '''expression : NOT expression
        '''
        p[0] = Not(p[2])

    def p_binary_operation(self, p):
        '''expression : expression POWER expression
                      | expression MULTIPLY expression
                      | expression DIVIDE expression
                      | expression MODULO expression
                      | expression INT_DIVIDE expression
                      | expression ADD expression
                      | expression MINUS expression
                      | expression EQUAL expression
                      | expression NOT_EQUAL expression
                      | expression LESS_EQUAL expression
                      | expression GREATER_EQUAL expression
                      | expression LESS expression
                      | expression GREATER expression
                      | expression AND expression
                      | expression OR expression
        '''
        p[0] = Operator.get_binary(p[2])(p[1], p[3])

    def p_parenthesis(self, p):
        '''expression : LEFT_PARENTHESIS expression RIGHT_PARENTHESIS
        '''
        p[0] = ParenthesisGroup(p[2])

    def p_if(self, p):
        '''expression : expression QUESTION expression COLON expression
        '''
        p[0] = If(p[1], p[3], p[5])

    def p_brace_expression(self, p):
        '''brace_expression : LEFT_BRACE expression RIGHT_BRACE
        '''
        p[0] = BraceGroup(p[2])

    def p_brace(self, p):
        '''expression : brace_expression
        '''
        p[0] = p[1]

    def p_quote(self, p):
        '''expression : QUOTE expression QUOTE
        '''
        p[0] = QuoteGroup(p[2])

    def p_expression_list_space(self, p):
        '''expression_list_space : item
                                | expression_list_space item
        '''
        if len(p) == 2:
            p[0] = SpaceList(p[1])
        else:
            p[1].append(p[2])
            p[0] = p[1]
    def p_item(self, p):
        '''item : expression
                | tuple
        '''
        p[0] = p[1]


    def p_expression_list_comma(self, p):
        '''expression_list_comma : expression
                                 | expression_list_comma COMMA expression
        '''
        if len(p) == 4:
            p[1].append(p[3])
            p[0] = p[1]
        else:
            p[0] = CommaList(p[1])

    def p_array(self, p):
        '''expression : LEFT_BRACKET expression_list_space RIGHT_BRACKET'''
        p[0] = BracketGroup(p[2])

    def p_tuple(self, p):
        '''tuple : LEFT_PARENTHESIS expression COMMA expression RIGHT_PARENTHESIS
        '''
        p[0] = Tuple(p[2], p[4])

    def p_tuple_list(self, p):
        '''tuple_list : tuple
                      | tuple_list tuple
        '''
        if len(p) == 3:
            p[1].append(p[2])
            p[0] = p[1]
        else:
            p[0] = SpaceList(p[1])

    def p_function(self, p):
        '''function : ID LEFT_PARENTHESIS expression_list_comma RIGHT_PARENTHESIS
                    | ID LEFT_PARENTHESIS expression_list_space RIGHT_PARENTHESIS
        '''
        # Match: sqrt(9)
        p[0] = Function(p[1], p[3])

    def p_function_expression(self, p):
        '''expression : function'''
        p[0] = p[1]

    def p_port_type_modifier(self, p):
        '''port_type_modifier : MODULO ID
        '''
        # Fixme: use token %[a-z]+ ?
        p[0] = PortTypeModifier(p[2])

    def p_port_modifier_function(self, p):
        '''expression : port_type_modifier LEFT_PARENTHESIS expression_list_space RIGHT_PARENTHESIS
        '''
        p[0] = PortModifierFunction(p[1], p[3])

    def p_port_modifier_vector(self, p):
        '''expression : port_type_modifier LEFT_BRACKET expression_list_space RIGHT_BRACKET
        '''
        p[0] = PortModifierVector(p[1], p[3])

    def p_set_id(self, p):
        '''expression : ID SET expression
                      | ID SET expression COMMA expression
        '''
        if len(p) == 5:
            p[0] = Set(Id(p[1]), Tuple(p[3], p[5]))
        else:
            p[0] = Set(Id(p[1]), p[3])

    def p_parameter(self, p):
        # Fixme:
        #   .ic v(cc) = 0 v(cc2) = 0
        #   Q23 10 24 13 QMOD IC=0.6, 5.0
        # Fixme: expression => recursive
        '''expression : function SET expression
                      | brace_expression SET tuple_list
        '''
        p[0] = Set(p[1], p[3])

    def p_error(self, p):
        if p is not None:
            self._logger.error(f"Syntax error at '{p.value}'")
            raise NameError('Syntax Error')
        else:
            self._logger.error("Syntax Error at End Of File")
            raise NameError("Syntax Error at End Of File")

    ##############################################

    def __init__(self) -> None:
        self._build()

    ##############################################

    def _build(self, **kwargs) -> None:
        self._lexer = lex.lex(module=self, **kwargs)
        self._parser = yacc.yacc(module=self, **kwargs)

    ##############################################

    def lex(self, text: str):
        self._lexer.input(text)
        while True:
            token = self._lexer.token()
            if token:
                yield token
            else:
                break

    ##############################################

    @staticmethod
    def dot_command(line: str) -> Optional[str]:
        if not line.startswith('.'):
            return None
        i = line.find(' ')
        if i == -1:
            _ = line
        else:
            _ = line[:i]
        return _.lower()

    @staticmethod
    def right_of(line: str, prefix: str) -> str:
        return line[len(prefix):].strip()

    ##############################################

    def parse(self, line: str) -> Command:
        """Parse a spice line.

        The LALR parser don't handle :code:`.include`, :code:`.lib` and :code:`.title`.

        It raises :code:`NameError('Syntax Error')`

        """

        line = line.strip()

        dot_command = self.dot_command(line)
        if dot_command is not None:
            match dot_command:
                case '.include':
                    return DotCommand(
                        '.include',
                        SpaceList(
                            Text(self.right_of(line, '.include').strip('"'))
                        )
                    )
                case '.lib':
                    # Fixme: how ngspice handle space in filename ?
                    _ = self.right_of(line, '.lib')
                    i = _.rfind(' ')
                    return DotCommand(
                        '.lib',
                        SpaceList(
                            Text(_[:i].rstrip()),
                            Text(_[i+1:])
                        )
                    )
                case '.title':
                    # Not handled by parser by commodity
                    return DotCommand(
                        '.title',
                        SpaceList(
                            Text(self.right_of(line, '.title'))
                        )
                    )

        # self._parser.defaulted_states = {}
        return self._parser.parse(line, lexer=self._lexer, debug=False)



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = []

####################################################################################################

import logging

from . import BasicElement
from . import HighLevelElement
from .Element import ElementParameterMetaClass
from .Netlist import Netlist

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

def _get_elements(module):
    return [
        _
        for _ in module.__dict__.values()
        if type(_) is ElementParameterMetaClass and _.PREFIX is not None
    ]

####################################################################################################

def _init():
    """Add a method to create elements to the Netlist class.

    .. code-block::

        circuit.R(*args, **kwargs)
        # =>
        R(circuit, *args, **kwargs)

    """

    spice_elements = _get_elements(BasicElement)
    high_level_elements = _get_elements(HighLevelElement)

    for element_class in spice_elements + high_level_elements:

        def make_wrapper(element_class):
            def function(self, *args, **kwargs):
                return element_class(self, *args, **kwargs)
            # Preserve docstrings for element shortcuts
            # Fixme: But Sphinx redumps it...
            function.__doc__ = element_class.__doc__
            function.ELEMENT_CLASS = element_class
            return function

        wrapper = make_wrapper(element_class)

        def register(name):
            # _module_logger.debug("Add device shortcut {} for class {}".format(name, element_class))
            setattr(Netlist, name, wrapper)

        register(element_class.__name__)
        if element_class in spice_elements:
            if hasattr(element_class, 'ALIAS'):
                register(element_class.ALIAS)
            if hasattr(element_class, 'LONG_ALIAS'):
                register(element_class.LONG_ALIAS)

_init()



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

"""This modules implements circuit and subcircuit.

The definition of a netlist follows the same conventions as SPICE. For example this SPICE netlist
is translated to Python like this:

.. code-block:: spice

    .title Voltage Divider
    Vinput in 0 10V
    R1 in out 9k
    R2 out 0 1k
    .end

.. code-block:: python3

    circuit = Circuit('Voltage Divider')
   circuit.V('input', 'in', circuit.gnd, 10)
    circuit.R(1, 'in', 'out', kilo(9))
    circuit.R(2, 'out', circuit.gnd, kilo(1))

or as a class definition:

.. code-block:: python3

      class VoltageDivider(Circuit):

          def __init__(self, **kwargs):

              super().__init__(title='Voltage Divider', **kwargs)

              self.V('input', 'in', self.gnd, '10V')
              self.R(1, 'in', 'out', kilo(9))
              self.R(2, 'out', self.gnd, kilo(1))

The circuit attribute :attr:`gnd` represents the ground of the circuit or subcircuit, usually set to
0.

We can get an element or a model using its name using these two possibilities::

    circuit['R1'] # dictionary style
    circuit.R1    # attribute style

The dictionary style always works, but the attribute only works if it complies with the Python
syntax, i.e. the element or model name is a valide attribute name (identifier), i.e. starting by a
letter and not a keyword like 'in', cf. `Python Language Reference
<https://docs.python.org/2/reference/lexical_analysis.html>`_.

We can update an element parameter like this::

    circuit.R1.resistance = kilo(1)

To simulate the circuit, we must create a simulator instance using the :meth:`Circuit.simulator`::

    simulator = circuit.simulator()

"""

####################################################################################################

from collections import OrderedDict
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, Self, Union
import keyword
import logging
import os

# import networkx

####################################################################################################

from InSpice.Tools.TextBuffer import TextBuffer
# Fixme: circular import
# from . import Library
from .DeviceModel import DeviceModel
from .Element import Pin, Element
from .StringTools import join_list, prefix_lines

if TYPE_CHECKING:
    from .Simulator import Simulator
    from . import Library

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class Node:

    """This class implements a node in the circuit.

    It stores a reference to the pins connected to the node.

    """

    _logger = _module_logger.getChild('Node')

    SPICE_GROUND_NUMBER = 0
    SPICE_GROUND_NAME = str(SPICE_GROUND_NUMBER)

    ##############################################

    @classmethod
    def _warn_iskeyword(cls, name: str) -> None:
        if keyword.iskeyword(name):
            cls._logger.warning(f"Node name '{name}' is a Python keyword")

    ##############################################

    def __init__(self, netlist: 'Netlist', name: str) -> None:
        self._warn_iskeyword(name)
        self._netlist = netlist
        self._name = str(name)
        self._pins = set()

    ##############################################

    def __repr__(self) -> str:
        return f'Node {self._name}'

    def __str__(self) -> str:
        return self._name

    ##############################################

    @property
    def netlist(self) -> 'Netlist':
        return self._netlist

    @property
    def name(self) -> int | str:
        return self._name

    @name.setter
    def name(self, value: int | str) -> bool:
        self._warn_iskeyword(value)
        self._name = value
        # update nodes dict
        self._netlist._update_node_name(self, value)

    @property
    def is_ground_node(self) -> bool:
        return self._name in (Node.SPICE_GROUND_NAME, 'gnd')

    ##############################################

    def __bool__(self) -> bool:
        return bool(self._pins)

    def __len__(self) -> int:
        return len(self._pins)

    def __iter__(self) -> Iterator[Pin]:
        return iter(self._pins)

    @property
    def pins(self) -> Iterator[Pin]:
        # Fixme: iter ?
        return iter(self._pins)

    def __contains__(self, pin: Pin) -> bool:
        return pin in self._pins

    ##############################################

    def connect(self, pin: Pin) -> None:
        self._logger.info(f"Connect {pin} => {self}")
        if pin not in self:
            self._pins.add(pin)
        else:
            # Fixme: could just warn ???
            raise ValueError(f"Pin {pin} is already connected to node {self}")

    ##############################################

    def disconnect(self, pin: Pin) -> None:
        self._logger.info(f"Disconnect {pin}")
        self._pins.remove(pin)

    ##############################################

    def merge(self, node: 'Node') -> None:
        self._logger.info(f"Merge {self} and {node}")
        for pin in list(node.pins):
            pin.disconnect()
            pin.connect(self)
        self._netlist._del_node(node)

    ##############################################

    def __iadd__(self, args: Union['Node', Pin, list[Union['Node', Pin]]]) -> Self:
        """Connect a node, a pin or a list of them to the node."""
        if isinstance(args, (Node, Pin)):
            args = (args,)
        for obj in args:
            if isinstance(obj, Node):
                # node <=> node
                self.merge(obj)
            elif isinstance(obj, Pin):
                # node <= pin
                if obj.connected:
                    self.merge(obj.node)
                else:
                    obj.connect(self)
            else:
                raise ValueError(f"Invalid object {type(obj)}")
        return self

####################################################################################################

class Netlist:

    """This class implements a base class for a netlist.

    .. note:: This class is completed with element shortcuts when the module is loaded.

    """

    _logger = _module_logger.getChild('Netlist')

    ##############################################

    def __init__(self) -> None:
        self._nodes = {}
        self._ground_name = Node.SPICE_GROUND_NAME   # Fixme: just here
        self._ground_node = self._add_node(self._ground_name)
        self._ground = None   # Fixme: purpose ???

        self._subcircuits = OrderedDict()   # to keep the declaration order
        self._elements = OrderedDict()   # to keep the declaration order
        self._models = {}

        self.raw_spice = ''

        # self._graph = networkx.Graph()

    ##############################################

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)

    ##############################################

    def copy_to(self, netlist: 'Netlist') -> 'Netlist':
        for subcircuit in self.subcircuits:
            netlist.subcircuit(subcircuit)
        for element in self.elements:
            element.copy_to(netlist)
        for name, model in self._models.items():
            netlist._models[name] = model.clone()
        netlist.raw_spice = str(self.raw_spice)
        return netlist

    ##############################################

    @property
    def gnd(self) -> int | str:
        # Fixme: purpose ???
        # return self._ground
        return self._ground_node

    # Note:
    #   circuit.gnd += ...
    #   call a setter...

    @property
    def nodes(self) -> Iterator[Node]:
        return self._nodes.values()

    @property
    def node_names(self) -> Iterator[str]:
        return self._nodes.keys()

    @property
    def elements(self) -> Iterator[Element]:
        return self._elements.values()

    @property
    def element_names(self) -> Iterator[str]:
        return self._elements.keys()

    @property
    def models(self) -> Iterator[DeviceModel]:
        return self._models.values()

    @property
    def model_names(self) -> Iterator[str]:
        return self._models.keys()

    @property
    def subcircuits(self) -> Iterator['SubCircuit']:
        return self._subcircuits.values()

    @property
    def subcircuit_names(self) -> Iterator[str]:
        return self._subcircuits.keys()

    ##############################################

    def element(self, name: str) -> Element:
        return self._elements[name]

    # Fixme: clash with
    #    def model(self, name, modele_type, **parameters):
    # def model(self, name):
    #     return self._models[name]

    # Fixme: versus get node ???
    def node(self, name: str) -> Node:
        return self._nodes[str(name)]

    ##############################################

    def __getitem__(self, attribute_name) -> Element:
        if attribute_name in self._elements:
            return self.element(attribute_name)
        elif attribute_name in self._models:
            # Fixme: error missing modele_type ?
            return self.model(attribute_name)
        # Fixme: subcircuits
        elif attribute_name in self._nodes:
            return self.node(attribute_name)
        else:
            raise IndexError(attribute_name)   # KeyError

    ##############################################

    def __getattr__(self, attribute_name):
        try:
            return self.__getitem__(attribute_name)
        except IndexError:
            raise AttributeError(attribute_name)

    ##############################################

    def _add_node(self, node_name: int | str) -> Node:
        node_name = str(node_name)
        if node_name not in self._nodes:
            self._logger.info(f'Create node "{node_name}"')
            node = Node(self, node_name)
            self._nodes[node_name] = node
            return node
        else:
            raise ValueError(f"Node {node_name} is already defined")

    ##############################################

    def _del_node(self, node) -> None:
        del self._nodes[node.name]

    ##############################################

    def _update_node_name(self, node, new_name) -> None:
        """Update the node's map for the new node's name"""
        # Fixme: check node is None ???
        if node.name not in self._nodes:
            # should not happen
            raise ValueError(f"Unknown node {node}")
        self._nodes[new_name] = self._nodes.pop(node.name)

    ##############################################

    def get_node(self, node: Node | int | str, create: bool = False) -> Node:
        """Return a node. `node` can be a node instance or node name.  A node is created if `create` is set
        and the node don't yet exist.

        """
        # Fixme: dangling...
        if node is None:
            return None
        # Fixme: always ok ???
        if isinstance(node, Node):
            return node
        else:
            str_node = str(node)
            if str_node in self._nodes:
                return self._nodes[str_node]
            elif create:
                return self._add_node(str_node)
            else:
                raise KeyError(f"Node {node} doesn't exists")

    ##############################################

    def has_ground_node(self) -> bool:
        """Test if ground node is connected"""
        return bool(self._ground_node)

    ##############################################

    def _add_element(self, element: Element) -> None:
        """Add an element."""
        if element.name not in self._elements:
            self._elements[element.name] = element
        else:
            raise NameError(f"Element name {element.name} is already defined")

    ##############################################

    def _remove_element(self, element: Element) -> None:
        try:
            del self._elements[element.name]
        except KeyError:
            raise NameError(f"Cannot remove undefined element {element}")

    ##############################################

    def model(self, name: str, modele_type: str, **parameters) -> DeviceModel:
        """Add a model."""
        _ = DeviceModel(name, modele_type, **parameters)
        if _.name not in self._models:
            self._models[_.name] = _
        else:
            raise NameError(f"Model name {name} is already defined")
        return _

    ##############################################

    def subcircuit(self, subcircuit: 'SubCircuit') -> None:
        """Add a sub-circuit."""
        # Fixme: subcircuit is a class
        self._subcircuits[str(subcircuit.name)] = subcircuit

    ##############################################

    def __str__(self) -> str:
        """ Return the formatted list of element and model definitions. """
        # Fixme: order ???
        netlist = TextBuffer()
        netlist += self._str_raw_spice()
        netlist += self._str_subcircuits()   # before elements
        netlist += self._str_elements()
        netlist += self._str_models()
        return str(netlist)

    ##############################################

    def _str_raw_spice(self) -> str:
        return self.raw_spice.rstrip()

    ##############################################

    def _str_subcircuits(self) -> list:
        # ensure list instead of odict_values
        return list(self.subcircuits)

    ##############################################

    def _str_elements(self) -> list:
        return [element for element in self.elements if element.enabled]

    ##############################################

    def _str_models(self) -> list:
        # ensure list instead of dict_values
        return list(self.models)

####################################################################################################

class SubCircuit(Netlist):

    """This class implements a sub-cicuit netlist."""

    ##############################################

    def __init__(self, name: str, *nodes, **kwargs) -> None:
        if len(set(nodes)) != len(nodes):
            raise ValueError(f"Duplicated nodes in {nodes}")

        super().__init__()

        self._name = str(name)
        self._external_nodes = nodes

        # Fixme: ok ?
        self._ground = kwargs.pop('ground', Node.SPICE_GROUND_NUMBER)

        self._parameters = kwargs

    ##############################################

    def clone(self, name: str = None) -> None:
        if name is None:
            name = self._name

        # Fixme: clone parameters ???
        kwargs = dict(self._parameters)
        kwargs['ground'] = self._ground

        subcircuit = self.__class__(name, list(self._external_nodes), **kwargs)
        self.copy_to(subcircuit)

    ##############################################

    @property
    def name(self) -> str:
        return self._name

    @property
    def external_nodes(self) -> list[Node]:
        return self._external_nodes

    @property
    def parameters(self) -> dict:
        """Parameters"""
        return self._parameters

    ##############################################

    def check_nodes(self) -> None:
        """Check for dangling nodes in the subcircuit."""
        nodes = self._external_nodes
        connected_nodes = set()
        for element in self.elements:
            connected_nodes.add(nodes & element.nodes)
        not_connected_nodes = nodes - connected_nodes
        if not_connected_nodes:
            raise NameError(f"SubCircuit Nodes {not_connected_nodes} are not connected")

    ##############################################

    def __str__(self) -> str:
        """Return the formatted subcircuit definition."""
        netlist = TextBuffer()
        nodes = join_list(self._external_nodes)
        parameters = join_list(['f{key}={value}'
                                for key, value in self._parameters.items()])
        netlist += '.subckt ' + join_list((self._name, nodes, parameters))
        netlist += super().__str__()
        netlist += '.ends ' + self._name
        return str(netlist)

####################################################################################################

class SubCircuitFactory(SubCircuit):

    NAME = None
    NODES = None

    ##############################################

    def __init__(self, **kwargs) -> None:
        super().__init__(self.NAME, *self.NODES, **kwargs)

####################################################################################################

class Circuit(Netlist):

    """This class implements a cicuit netlist.

    To get the corresponding Spice netlist use::

       circuit = Circuit()
       ...
       str(circuit)

    """

    _logger = _module_logger.getChild('Circuit')

    ##############################################

    def __init__(
        self,
        title: str,   # pylint issue
        ground: int | str = Node.SPICE_GROUND_NUMBER,   # Fixme: gnd = Node.SPICE_GROUND_NUMBER
        global_nodes: list[int | str] = (),
    ) -> None:
        super().__init__()

        self.title = str(title)
        self._ground = ground
        self._global_nodes = set(global_nodes)   # .global
        self._includes = []   # .include
        self._libs = []   # .lib, contains a (name, section) tuple
        self._parameters = {}   # .param

        # Fixme: not implemented
        #  .csparam
        #  .func
        #  .if

    ##############################################

    def clone(self, title: str = None) -> 'Circuit':
        if title is None:
            title = self.title

        circuit = self.__class__(title, self._ground, set(self._global_nodes))
        self.copy_to(circuit)

        for include in self._includes:
            circuit.include(include)
        for name, value in self._parameters.items():
            self.parameter(name, value)

        return circuit

    ##############################################

    @property
    def includes(self) -> Iterator[Path]:
        return iter(self._includes)

    ##############################################

    def include(self, path: Union[Path, str, 'Library.SubCircuit'], warn: bool = True) -> None:
        """Include a file."""
        # Fixme: str(path) ?
        # Fixme: circular import...
        from . import Library
        if isinstance(path, Library.Subcircuit) or isinstance(path, Library.Model):
            path = path.path
        path = Path(path).resolve()
        if path not in self._includes:
            self._includes.append(path)
        elif warn:
            self._logger.warn(f"Duplicated include {path}")

    ##############################################

    def lib(self, name: str, section: str = None) -> None:
        """Load a library."""
        v = (name, section)
        if v not in self._libs:
            self._libs.append(v)
        else:
            self._logger.warn(f"Duplicated lib {v}")

    ##############################################

    def parameter(self, name: str, expression: str) -> None:
        """Set a parameter."""
        self._parameters[str(name)] = str(expression)

    ##############################################

    def str(self, simulator: 'Simulator' = None) -> str:
        """Return the formatted desk.

        :param simulator: simulator instance to select the flavour of a Spice library

        """
        # if not self.has_ground_node():
        #     raise NameError("Circuit don't have ground node")
        _ = TextBuffer()
        _ += self._str_title()
        _ += self._str_includes(simulator)
        _ += self._str_libs(simulator)
        _ += self._str_globals()
        _ += self._str_parameters()
        _ += super().__str__()
        return str(_) + os.linesep    # Fixme: linesep here ???

    ##############################################

    def _str_title(self) -> str:
        return f'.title {self.title}'

    ##############################################

    def _str_includes(self, simulator: 'Simulator' = None) -> list[str]:
        if self._includes:
            # ngspice don't like // in path, thus ensure we write real paths
            real_paths = []
            for path in self._includes:
                if simulator:
                    path_flavour = path.parent.joinpath(f"{path.name}@{simulator}")
                    if path_flavour.exists():
                        path = path_flavour
                real_paths.append(path)
            return prefix_lines(real_paths, prefix='.include ')
        else:
            return None

    ##############################################

    def _str_libs(self, simulator: 'Simulator' = None) -> list[str]:
        if self._libs:
            libs = []
            for lib, section in self._libs:
                lib = Path(str(lib)).resolve()
                if simulator:
                    lib_flavour = Path(f"{lib}@{simulator}")
                    if lib_flavour.exists():
                        lib = lib_flavour
                s = f".lib {lib}"
                if section:
                    s += f" {section}"
                libs.append(s)
            return libs
        else:
            return None

    ##############################################

    def _str_globals(self) -> str:
        if self._global_nodes:
            return '.global ' + join_list(self._global_nodes)
        else:
            return None

    ##############################################

    def _str_parameters(self) -> list[str]:
        if self._parameters:
            return [f'.param {key}={value}' for key, value in self._parameters.items()]
        else:
            return None

    ##############################################

    def __str__(self) -> str:
        return self.str(simulator=None)

    ##############################################

    def str_end(self) -> str:
        return str(self) + '.end'

    ##############################################

    def simulator(self, *args, **kwargs):
        # return CircuitSimulator.factory(self, *args, **kwargs)
        raise NameError("Deprecated API")



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

from pathlib import Path
import argparse

####################################################################################################

from InSpice.Logging import Logging
logger = Logging.setup_logging()

####################################################################################################

from InSpice.Spice.Library import SpiceLibrary, Model, Subcircuit

####################################################################################################

def main() -> None:
    parser = argparse.ArgumentParser(description='Manage Spice library')
    parser.add_argument(
        '-l',
        '--library_path',
        required=True,
        help='Spice library root path',
    )
    parser.add_argument(
        '--scan',
        default=False,
        action='store_true',
        help='Scan library',
    )
    parser.add_argument(
        '--delete-yaml',
        default=False,
        action='store_true',
        help='WARNING: Delete YAML Files',
    )
    parser.add_argument(
        '--add-category',
        type=str,
        help='Add a category',
    )
    parser.add_argument(
        '--category',
        type=str,
        default=None,
        help='specify the category',
    )
    parser.add_argument(
        '--add-library',
        type=str,
        help='Add a library',
    )
    parser.add_argument(
        '--show-categories',
        default=False,
        action='store_true',
        help='Show categories',
    )
    parser.add_argument(
        '--search',
        help='search',
    )
    args = parser.parse_args()

    ##############################################

    library_path = Path(args.library_path).resolve()
    print(f"Library is {library_path}")

    spice_library = SpiceLibrary(library_path, scan=args.scan)

    if args.delete_yaml:
        rc = input('Confirm deletion (y/n): ')
        if rc.lower() == 'y':
            spice_library.delete_yaml()

    if args.show_categories:
        print(spice_library.list_categories())

    if args.add_category:
        spice_library.category_path(args.add_category)

    if args.add_library:
        if args.category is None:
            print("A category is required")
        raise NotImplementedError

    if args.search:
        print()
        print(f'Results for "{args.search}":')
        # list() to don't mix logging and print
        for name, include in list(spice_library.search(args.search)):
            path = include.path.relative_to(library_path)
            print(' '*2 + f"{name} {path}")
            indent = ' '*4
            if include.description:
                print(f"{indent}Library: {include.description}")
            _ = include[name]
            is_model = isinstance(_, Model)
            if is_model:
                print(f"{indent}Model: {_.name} — {_.description}")
            else:
                print(f"{indent}Subcircuit: {_.name} — {_.description}")
                pins = ' '.join(_.pin_names)
                print(f"{indent}Pins: {pins}")



####################################################################################################
from pathlib import Path

from KiCadRW.Schema import KiCadSchema
from KiCadRW.Drawings.CircuitMacros import CircuitMacrosDumper
from InSpice.KiCad import PythonDumper

####################################################################################################

schema_path = Path(
    'examples', 'power-supplies', 'kicad', 'capacitive-half-wave-rectification-pre-zener',
    'capacitive-half-wave-rectification-pre-zener.kicad_sch'
)

kicad_schema = KiCadSchema(schema_path)
print()
print('='*100)
kicad_schema.dump_netlist()

print()
print('='*100)
python_code = PythonDumper(kicad_schema, use_InSpice_unit=True)
print(python_code)

print()
print('='*100)
cm_code = CircuitMacrosDumper(kicad_schema)
print(cm_code)



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

# Simulation outputs should be identical if
#   same simulator and version
#   same inputs

__all__ = ["SimulationCache"]

####################################################################################################

from pathlib import Path
import hashlib

# http://www.grantjenks.com/docs/diskcache
from diskcache import Cache

####################################################################################################

class SpiceInclude:

    INCLUDE_PREFIX = '.include '

    ##############################################

    def __init__(self, path: str | Path):
        self._path = Path(path)
        self._inner_includes = []
        self._walk()
        self._digest = self.sha1()
        self._recursive_digest = self._compute_recursive_digest()

    ##############################################

    @property
    def path(self) -> Path:
        return self._path

    ##############################################

    def _compute_digest(self, func) -> str:
        with open(self._path, 'rb') as fh:
            _ = func(fh.read()).hexdigest()
        return _

    ##############################################

    def sha1(self) -> str:
        return self._compute_digest(hashlib.sha1)

    ##############################################

    @classmethod
    def parse_include(self, line):
        line = line.strip()
        if line.startswith(self.INCLUDE_PREFIX):
            path = line[len(self.INCLUDE_PREFIX):].strip()
            if path:
                return Path(path)
        return None

    ##############################################

    def _walk(self) -> None:
        with open(self._path, 'r') as fh:
            for line in fh.readlines():
                path = self.parse_include(line)
                if path is not None:
                    include = SpiceInclude(path)
                    self._inner_includes.append(include)

    ##############################################

    def _compute_recursive_digest(self) -> str:
        if self._inner_includes:
            return self._digest + '[' + '/'.join([_.digest for _ in self._inner_includes]) + ']'
        else:
            return self._digest

    ##############################################

    @property
    def digest(self) -> str:
        return self._digest

    @property
    def recursive_digest(self) -> str:
        return self._recursive_digest

    @property
    def inner_includes(self) -> str:
        return iter(self._inner_includes)

####################################################################################################

class SimulationCache:

    ##############################################

    def __init__(self, path: str | Path=None):
        """
        Cache directory will be set to a temporary directory like "/tmp/diskcache-...", if *Path* is None.
        """
        self._cache = Cache(path)

    ##############################################

    def __del__(self):
        # Each thread that accesses a cache should also call close on the cache.
        self._cache.close()

    ##############################################

    @property
    def impl(self) -> Cache:
        return self._cache

    @property
    def directory(self) -> str:
        return self._cache.directory

    ##############################################

    def simulation_key(self, simulation) -> str:
        raw_spice_code = str(simulation)
        lines = []
        # Replace .include lines by recursive digest
        for line in raw_spice_code.splitlines():
        #  for path in simulation.circuit.includes:
            path = SpiceInclude.parse_include(line)
            if path is not None:
                include = SpiceInclude(path)
                lines.append(f'.include {include.path.name} {include.recursive_digest}')
            else:
                lines.append(line)
        # Replace os.linesep to a portable separator
        spice_code = '@@'.join(lines)
        return f'{simulation.simulator_name}/{simulation.simulator_version}/{spice_code}'

    ##############################################

    def add(self, analysis):
        pass

    ##############################################

    def get(self, simulation):
        pass

####################################################################################################

class CachedSimulation:

    ##############################################

    def __init__(self, spice_code):
        pass



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2017 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = ['find_libraries', 'LIBRARY_PATH']

####################################################################################################

from pathlib import Path
import logging
import os
import sys

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

LIBRARY_PATH = 'spice-library'

def find_libraries(root: str='examples') -> Path:
    try:
        library_path = os.environ['InSpiceLibraryPath']
    except KeyError:
        examples_root = Path(sys.argv[0]).resolve()   # path of the Python file
        while True:
            examples_root = examples_root.parents[1]
            if examples_root.name == root:
                break
        library_path = examples_root.joinpath(LIBRARY_PATH)
    _module_logger.info(f'SPICE library path is {library_path}')
    return library_path



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module implements a Spice parser.

"""

####################################################################################################

__all__ = [
    'Ac',
    'Command',
    'Control',
    'Csparam',
    'Dc',
    'Distorsion',
    'DotCommand',
    'Element',
    'Else',
    'ElseIf',
    'EndControl',
    'End',
    'EndIf',
    'EndSubcircuit',
    'Fourier',
    'Function',
    'Global',
    'If',
    'Include',
    'InitialCondition',
    'Library',
    'Measure',
    'Model',
    'Netlist',
    'NodeSet',
    'Noise',
    'Op',
    'Options',
    'Param',
    'ParseError',
    'PeriodicSteadyState',
    'Plot',
    'PoleZero',
    'Print',
    'Probe',
    'Save',
    'Sensitivity',
    'SpiceSource',
    'SpiceFile',
    'SpiceLine',
    'SpiceStates',
    'Subcircuit',
    'Temperature',
    'Title',
    'Transient',
    'Width',
]

####################################################################################################

from enum import IntEnum
from pathlib import Path
from typing import Generator, Iterator, Optional
import logging
import os

from ..Netlist import Node
from ..StringTools import remove_multi_space
from . import Ast
from .ElementData import ElementData
from .Ast import AstNode
from .Parser import SpiceParser
from .SpiceSyntax import ElementLetters

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class ParseError(NameError):
    pass

####################################################################################################

class SpiceLine:

    _logger = _module_logger.getChild('SpiceLine')

    ##############################################

    def __init__(self, start: int, stop: int, command: str, comment: str) -> None:
        self._start = start
        self._stop = stop
        self._command = command
        self._comment = comment
        self._is_dot_command = None
        self._is_element = None
        self._dot_command = None

    ##############################################

    @property
    def start(self) -> int:
        return self._start

    @property
    def location(self) -> slice:
        return slice(self._start, self._stop)

    @property
    def str_location(self) -> str:
        return f'[{self._start}:{self._stop}]'

    @property
    def command(self) -> str:
        return self._command

    @property
    def comment(self) -> str:
        return self._comment

    ##############################################

    @property
    def is_comment(self) -> bool:
        return not self._command

    @property
    def is_command(self) -> bool:
        return bool(self._command)

    @property
    def is_dot_command(self) -> bool:
        if self._is_dot_command is None:
            self._is_dot_command = self._command and self._command.startswith('.')
        return self._is_dot_command

    @property
    def is_element(self) -> bool:
        if self._is_element is None:
            self._is_element = self.is_command and not self.is_dot_command
        return self._is_element

    ##############################################

    @property
    def dot_command(self) -> Optional[str]:
        if not self.is_dot_command:
            return None
        if self._dot_command is None:
            i = self._command.find(' ')
            if i == -1:
                _ = self._command
            else:
                _ = self._command[:i]
            self._dot_command = _.lower()
        return self._dot_command

    # @property
    # def element_letter(self) -> str:
    #     if self.is_element:
    #         return self._command[0].upper()
    #     else:
    #         raise ValueError

    ##############################################

    def append(self, line_number: int, command: str | None, comment: str | None) -> None:
        self._stop = line_number
        if command:
            self._command += ' ' + command
        if comment:
            self._comment += os.linesep + comment

    ##############################################

    def cleanup(self) -> None:
        """Remove multi-space"""
        # Fixme: tab ???
        self._command = remove_multi_space(self._command)

    ##############################################

    def __repr__(self) -> str:
        line_number = self.str_location
        if self.is_comment:
            return f'{line_number} COMMENT{os.linesep}    {self._comment}'
        else:
            if self.is_dot_command:
                label = 'DOT COMMAND'
            else:
                label = 'ELEMENT / CONTROL COMMAND'
            return f'{line_number} {label}{os.linesep}    {self._command} ; {self._comment}'

    ##############################################

    # def right_of(self, prefix: str) -> str:
    #     return self._command[len(prefix):].strip()

####################################################################################################

class Command:

    DOT_COMMAND = None
    _dot_command_maps = {}

    ##############################################

    def __init_subclass__(cls, **kwargs) -> None:
        if cls.DOT_COMMAND is not None:
            cls._dot_command_maps[cls.DOT_COMMAND] = cls

    ##############################################

    @classmethod
    def get_cls(cls, line: SpiceLine, in_control: bool) -> 'Command':
        # Fixme: use AST instead ?
        if not in_control and line.is_element:
            return Element
        elif in_control or line.is_dot_command:
            try:
                return cls._dot_command_maps[line.dot_command]
            except KeyError:
                return cls
        return None

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        self._line = line
        self._ast = ast

    ##############################################

    @property
    def line(self) -> SpiceLine:
        return self._line

    @property
    def ast(self) -> AstNode:
        return self._ast

    ##############################################

    @staticmethod
    def format_parameters(parameters: dict) -> dict[str, str]:
        return {key: str(value) for key, value in parameters.items()}

    ##############################################

    @staticmethod
    def child_to_python(child: AstNode) -> str | int:
        match child:
            case Ast.Id():
                return str(child)
            case Ast.Integer():
                return int(child)
            case Ast.Number():
                if child.has_str:
                    return str(child)
                else:
                    return float(child)
            case _:
                # Fixme:
                return str(child)

    @classmethod
    def childs_to_python(cls, childs: list[AstNode]) -> list:
        return [cls.child_to_python(_) for _ in childs]

####################################################################################################

class Element(Command):

    """ This class implements an element definition.

    Spice syntax::

        [A-Z][name] ...

    """

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        super().__init__(line, ast)

        def is_linear_element(command_ast):               
            for child in command_ast:
                if not isinstance(child, (Ast.Id, Ast.Integer, Ast.Number)):
                    return False   # found something complicated (like Function, Set, etc)
            return True

        self._letter = ast.first_letter
        if not getattr(ElementLetters, self._letter):
            raise ParseError(f"Invalid element letter in element command @{line.str_location} {self._ast.name}")
        self._name = ast.after_first_letter

        first_set_position = ast.childs.find_first_set()

        # Read nodes
        self._nodes = []
        number_of_pins = 0
        from InSpice.Spice.Element import ElementParameterMetaClass
        elements = {}
        for letter, classes in ElementParameterMetaClass._classes.items():
            element_data = ElementData(letter, classes)
            elements[letter] = element_data
            elements[letter.lower()] = element_data
        # data = ElementData.elements[self._letter]
        data = elements[self._letter]
        number_of_positional_parameters = None
        migh_have_dict_parameters = True
        if not data.has_variable_number_of_pins:
            if data.multi_devices and not is_linear_element(ast):  # maybe redundant but  okay for now. 
                number_of_pins = len(data.classes[1].PINS)  # Nonlinear E, G, H, F, etc
                number_of_positional_parameters = data.classes[1].number_of_positional_parameters
                migh_have_dict_parameters = False
            else:
                number_of_pins = data.number_of_pins
        else:   # Q or X
            if first_set_position == -1:
                number_of_pins = len(ast)
            else:
                number_of_pins = first_set_position
            number_of_pins -= 1   # last is model / subcircuit name
            if self._letter == 'Q':
                last_node = ast[number_of_pins -1]
                if isinstance(last_node, Ast.Id) and str(last_node) == 'off':
                    number_of_pins -= 1
                if not (3 <= number_of_pins <= 5):
                    raise ValueError("Invalid number of nodes for {ast}")
            self._model_name = str(ast[number_of_pins])
        if number_of_pins:
            self._nodes = self.childs_to_python(ast[:number_of_pins])

        self._parameters = []
        self._dict_parameters = {}

        for child in ast[number_of_pins:]:
            match child:
                case Ast.Id():
                    self._parameters.append(str(child))
                case Ast.Integer():
                    self._parameters.append(int(child))
                # case Ast.Number():
                #     self._parameters.append(child)
                case Ast.Set():
                    if migh_have_dict_parameters:
                        self._dict_parameters[child.left_id] = child
                    else:
                        self._parameters.append(child)    # in the case of Table (expression) = {...}
                case _:
                    self._parameters.append(child)
                    # raise ValueError(f"Invalid item {child} in {ast}")

        # Fixme: ok ???
        # in the case of nonlinear element with multiple devices the self._parameters are not equal to the number of positional parameters
        if data.multi_devices:
            if number_of_positional_parameters is None:
                number_of_positional_parameters = len(self._parameters)
            for element_class in data:
                if number_of_positional_parameters == element_class.number_of_positional_parameters:
                    break
        else:
            element_class = data.single
        self._factory = element_class

    ##############################################

    @property
    def factory(self):
        return self._factory

    @property
    def letter(self) -> str:
        return self._letter

    @property
    def name(self) -> str:
        return self._name

    @property
    def model_name(self) -> str:
        if hasattr(self, 'model_name'):
            return self._model_name
        else:
            return None

    subcircuit_name = model_name

    @property
    def parameters(self):
        return self._parameters

    @property
    def dict_parameters(self):
        return self._dict_parameters

    ##############################################

    def __repr__(self) -> str:
        parameters = [str(_) for _ in self._parameters]
        dict_parameters = self.format_parameters(self._dict_parameters)
        return f'Element {self._letter}[{self._name}] {self._nodes} {parameters} {dict_parameters}'

    ##############################################

    def translate_ground_node(self, ground: int) -> list[str|int]:
        nodes = []
        for node in self._nodes:
            if str(node) == str(ground):
                node = Node.SPICE_GROUND_NUMBER
            nodes.append(node)
        return nodes

####################################################################################################

class DotCommand(Command):
    pass

####################################################################################################

class Ac(DotCommand):
    DOT_COMMAND = '.ac'

####################################################################################################

class Control(DotCommand):
    DOT_COMMAND = '.control'

####################################################################################################

class Csparam(DotCommand):

    """This class implements a csparam statement.

    Create a constant vector from a parameter in plot const.

    Spice syntax::

        .CSPARAM <ident> = { <expr > }

    Examples::

        .PARAM pippo=5
        .PARAM pp=6
        .CSPARAM pippp={pippo + pp}
        .PARAM p={pp}
        .CSPARAM pap='pp+p'
    """

    DOT_COMMAND = '.param'

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        super().__init__(line, ast)
        # Fixme:

    ##############################################

    def __repr__(self) -> str:
        return f'Csparam'

####################################################################################################

class Dc(DotCommand):
    DOT_COMMAND = '.dc'

class Distorsion(DotCommand):
    DOT_COMMAND = '.disto'

class Else(DotCommand):
    DOT_COMMAND = '.else'

class ElseIf(DotCommand):
    DOT_COMMAND = '.elseif'

####################################################################################################

class End(DotCommand):
    DOT_COMMAND = '.end'

class EndControl(DotCommand):
    DOT_COMMAND = '.endc'

class EndIf(DotCommand):
    DOT_COMMAND = '.endif'

class EndSubcircuit(DotCommand):
    DOT_COMMAND = '.ends'

####################################################################################################

class Fourier(DotCommand):
    DOT_COMMAND = '.fourier'

####################################################################################################

class Function(DotCommand):

    """This class implements a func statement.

    This command defines a function. The syntax of the expression is the same as for a *.PARAM*.

    Spice syntax::

        .FUNC <ident> { <expr > }
        .FUNC <ident> = { <expr > }

    Examples::

        .FUNC icos(x) {cos(x) - 1}
        .FUNC f(x,y) {x*y}
        .FUNC foo(a,b) = {a + b}
    """

    DOT_COMMAND = '.func'

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        super().__init__(line, ast)
        # Fixme:
        self._name = ''
        self._variables = ''
        self._expression = ''

    ##############################################

    @property
    def name(self) -> str:
        return self._name

    @property
    def variables(self) -> list[str]:
        return self._variables

    @property
    def expression(self) -> str:
        return self._expression

    ##############################################

    def __repr__(self) -> str:
        return f'Func {self._name} {self._variables} {self._expression}'

####################################################################################################

class Global(DotCommand):

    """This class implements a global command.

    Spice syntax::

        .GLOBAL nodename
        .GLOBAL nodename1 nodename2 ...

    Examples::

        .GLOBAL gnd vcc
    """

    DOT_COMMAND = '.global'

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        super().__init__(line, ast)
        # Fixme
        self._nodes = ()

    ##############################################

    @property
    def nodes(self):   # ->
        return iter(self._nodes)

    ##############################################

    def __repr__(self) -> str:
        return f'Global {self._nodes}'

####################################################################################################

class InitialCondition(DotCommand):
    DOT_COMMAND = '.ic'

####################################################################################################

class If(DotCommand):
    DOT_COMMAND = '.if'

####################################################################################################

class Include(DotCommand):

    """This class implements a include command.

    Spice syntax::

        .INCLUDE filename

    Examples::

        .INCLUDE /users/spice/common/bsim3-param.mod
    """

    DOT_COMMAND = '.include'

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        super().__init__(line, ast)
        self._path = Path(str(ast.child))

    ##############################################

    @property
    def path(self) -> Path:
        return self._path

    ##############################################

    def __repr__(self) -> str:
        return f'Include "{self._path}"'

####################################################################################################

class Library(DotCommand):

    """This class implements a library command.

    Spice syntax::

        .LIB filename libname

    Examples::

        .LIB /users/spice/common/mosfets.lib mos1
    """

    DOT_COMMAND = '.lib'

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        super().__init__(line, ast)
        self._path, self._libname = [str(_) for _ in ast]
        self._path = Path(self._path)

    ##############################################

    @property
    def path(self) -> Path:
        return self._path

    @property
    def libname(self) -> str:
        return self._libname

    ##############################################

    def __repr__(self) -> str:
        return f'Lib "{self._path}" {self._libname}'

####################################################################################################

class Measure(DotCommand):
    DOT_COMMAND = '.meas'

####################################################################################################

class Model(DotCommand):

    """This class implements a model command.

    Spice syntax::

        .MODEL mname type (pname1=pval1 pname2=pval2)

    Examples::

        .MODEL MOD1 npn (bf=50 is=1e-13 vbf=50)
    """

    DOT_COMMAND = '.model'

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        super().__init__(line, ast)
        self._name = str(ast.first_child)
        child2 = ast[1]
        if isinstance(child2, Ast.Function):
            self._type = child2.name
            parameters = child2   # .childs
        else:
            self._type = str(child2)
            parameters = ast[2:]
        self._dict_parameters = {}
        # Fixme: duplicated code
        for child in parameters:
            if isinstance(child, Ast.Set):
                self._dict_parameters[child.left_id] = child.right
            else:
                raise ParseError(f"Invalid item '{child}' in {ast}")

    ##############################################

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> str:
        return self._type

    @property
    def parameters(self) -> dict:
        return self._dict_parameters

    @property
    def py_parameters(self) -> dict:
        return {name:self.child_to_python(ast) for name, ast in self._dict_parameters.items()}

    ##############################################

    def __repr__(self) -> str:
        parameters = self.format_parameters(self._dict_parameters)
        return f'Model {self._name} type={self._type} {parameters}'

####################################################################################################

class NodeSet(DotCommand):
    DOT_COMMAND = '.nodeset'

class Noise(DotCommand):
    DOT_COMMAND = '.noise'

class Op(DotCommand):
    DOT_COMMAND = '.op'

class Options(DotCommand):
    DOT_COMMAND = '.options'

####################################################################################################

class Param(DotCommand):

    """This class implements a param command.

    Spice syntax::

        .PARAM <ident> = <expr> <ident> = <expr> ...

    Examples::

        .PARAM pippo=5
        .PARAM po=6 pp=7.8 pap={AGAUSS(pippo, 1, 1.67)}
        .PARAM pippp={pippo + pp}
        .PARAM p={pp}
        .PARAM pop='pp+p'
    """

    DOT_COMMAND = '.param'

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        super().__init__(line, ast)
        # Fixme:
        self._parameters = {}

    ##############################################

    @property
    def parameters(self) -> dict:
        return self._parameters

    ##############################################

    def __repr__(self) -> str:
        return f'Param {self._parameters}'

####################################################################################################

class Plot(DotCommand):
    DOT_COMMAND = '.plot'

class Print(DotCommand):
    DOT_COMMAND = '.print'

class Probe(DotCommand):
    DOT_COMMAND = '.probe'

class PeriodicSteadyState(DotCommand):
    DOT_COMMAND = '.pss'

class PoleZero(DotCommand):
    DOT_COMMAND = '.pz'

class Save(DotCommand):
    DOT_COMMAND = '.save'

class Sensitivity(DotCommand):
    DOT_COMMAND = '.sens'

####################################################################################################

class Netlist:

    ##############################################

    def __init__(self) -> None:
        self._items = []

    ##############################################

    def add(self, item: Command) -> None:
        self._items.append(item)

    append = add

    ##############################################

    def __len__(self) -> int:
        return len(self._items)

    def __bool__(self) -> bool:
        return bool(self._items)

    def __iter__(self) -> Iterator[Command]:
        return iter(self._items)

####################################################################################################

class Subcircuit(Netlist, DotCommand):

    """This class implements a subcircuit command.

    Spice syntax::

        .SUBCKT name N1 <N2 ...>
        .SUBCKT name N1 <N2 ...>  <ident>=<value> <ident>=<value> ...

    Examples::

        .SUBCKT vdivide 1 2 3
        R1 1 2 10K
        R2 2 3 5
        .ENDS

       .SUBCKT myfilter in out rval=100k cval=100nF

    Each :code:`<value>` is either a SPICE number or a brace expression :code:`{<expr>}`.
    """

    DOT_COMMAND = '.subckt'

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        DotCommand.__init__(self, line, ast)
        Netlist.__init__(self)
        # self._name = str(ast.first_child)
        self._name, *self._nodes = self.childs_to_python(ast.childs.iter_on_leaf())
        self._dict_parameters = {}
        for child in ast[len(self._nodes)+1:]:
            if isinstance(child, Ast.Set):
                self._dict_parameters[child.left_id] = child.right
            else:
                raise ParseError(f"Invalid item {child} in {ast}")

    ##############################################

    @property
    def name(self) -> str:
        return self._name

    @property
    def nodes(self) -> list[str]:
        return self._nodes

    @property
    def parameters(self) -> dict:
        return self._dict_parameters

    ##############################################

    def __repr__(self) -> str:
        parameters = self.format_parameters(self._dict_parameters)
        return f'Subckt {self._name} {self._nodes} {parameters}'

####################################################################################################

class Temperature(DotCommand):

    """This class implements a temp command.

    Sets the circuit temperature in degrees Celsius.

    Spice syntax::

        .TEMP value

    Examples::

        .TEMP 27
    """

    COMMAND = '.temp'

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        super().__init__(line, ast)
        self._temperature = ast.child

    ##############################################

    @property
    def float_temperature(self) -> float:
        # Fixme: could fail
        return float(self._temperature)

    ##############################################

    def __repr__(self) -> str:
        return f'Temp {self._temperature}'

####################################################################################################

class Title(DotCommand):

    """This class implements a title command."""

    DOT_COMMAND = '.title'

    ##############################################

    def __init__(self, line: SpiceLine, ast: AstNode) -> None:
        super().__init__(line, ast)
        self._title = str(ast.child)

    ##############################################

    def __str__(self) -> str:
        return self._title

    ##############################################

    def __repr__(self) -> str:
        return f'Title "{self._title}"'

####################################################################################################

class Transient(DotCommand):
    DOT_COMMAND = '.tran'

class Width(DotCommand):
    DOT_COMMAND = '.width'

####################################################################################################

class SpiceStates(IntEnum):
    DESK = 0         # -> .end
    SUBCIRCUIT = 1   # -> .subcircuit -> .ends
    IF = 2           # -> .if -> .endif
    CONTROL = 3      # -> .control -> .endc / or .end

####################################################################################################

class SpiceSource:

    _logger = _module_logger.getChild('SpiceSource')

    ##############################################

    def __init__(self, source: Optional[str]=None, title_line: bool=True) -> None:
        self._parser = SpiceParser()
        self.reset()
        if source is not None:
            self.parse_string(source, title_line)

    ##############################################

    def reset(self) -> None:
        self._lines = []
        self._title_line = None
        self._ast_lines = []
        self._obj_lines = []
        self._titles = []
        self._includes = []
        self._libs = []
        self._models = []
        self._subcircuits = []
        self._circuit = Netlist()
        self._control = []
        self._section = None

    ##############################################

    @classmethod
    def _split_command_comment(cls, line):
        # Handle end-of-line comments
        command = ''
        comment = ''
        if '$' in line:
            command, comment = line.split('$')
        elif ';' in line:
            command, comment = line.split(';')
        else:
            command = line
        return command.strip(), comment.strip()

    ##############################################

    def read(self, generator: Generator[tuple[int, str], None, None], title_line: bool=True, section: str=None ) -> None:
        """Preprocess lines. This method merges continuation lines and split command and comment.

        """
        self.reset()
        last_line = None
        last_command = None
        we_are_in_lib = False
        self._section = section
        libname = None
        for line_number, line in generator:
            # print(f'>>>{line_number}///{line.rstrip()}')
            ### line = line.strip()
            # handle first line
            line_is_comment = line.startswith('*')
            if line_number == 0 and not line_is_comment and title_line:
                self._title_line = SpiceLine(line_number, line_number, None, line.strip())
                self._lines.append(self._title_line)
                continue
            # Skip empty line
            if not line.strip():
                continue
            command = ''
            comment = ''
            # Handle continuation line
            if line.startswith('+'):
                _ = line[1:].strip()
                command, comment = self._split_command_comment(_)
                if last_command:
                    last_command.append(line_number, command, comment)
                    continue
                else:
                    raise ParseError(f"Continuation line at {line_number} doesn't follow a command line")
            # Handle comment line
            elif line_is_comment:
                comment = line[1:].strip()
            else:
                command, comment = self._split_command_comment(line)
            # print(f'>>>{line_number}///{command}///{comment}///')
            # Is continuing comment ?
            if not command and last_line and last_line.is_comment:
                last_line.append(line_number, '', comment)
            else:
                last_line = SpiceLine(line_number, line_number, command, comment)
                if command.startswith('.lib'):
                    libname = command.split('lib')[1].strip().lower()
                    we_are_in_lib = True
                elif command.startswith('.endl'):
                    we_are_in_lib = False
                    libname = None
                if self._section and libname != self._section.lower() and we_are_in_lib:
                    continue
                self._lines.append(last_line)
                if command:
                    last_command = last_line

    ##############################################
    def _normalize_quotes(self, text: str) -> str:
        """Normalize Unicode quotes to ASCII quotes."""
        text = str(text)
        return text.translate(str.maketrans({
            '‘': "'", 
            '’': "'",
            '“': '"',
            '”': '"'
            }))
    
    def parse(self) -> None:

        for line in self._lines:
            if line.is_comment:
                self._ast_lines.append(None)
            else:
                line.cleanup()
                self._logger.debug(os.linesep + str(line))
                try:
                    ast = self._parser.parse(line.command)
                    self._ast_lines.append(ast)
                    self._logger.debug(os.linesep + str(ast.pretty_print()))
                except NameError as e:
                    raise ParseError(str(e))

    ##############################################

    def analyse(self) -> None:
        state_stack = [SpiceStates.DESK]
        circuit = []
        subcircuit = None
        control = []

        def get_state() -> int:
            return state_stack[-1]

        def not_state(state: int) -> bool:
            return get_state() != state

        def append(obj: Command) -> None:
            match get_state():
                case SpiceStates.SUBCIRCUIT:
                    _ = subcircuit
                case SpiceStates.CONTROL:
                    _ = control
                case _:
                    _ = circuit
            _.append(obj)

        for line, ast in zip(self._lines, self._ast_lines):
            if line.is_comment:
                continue
            cls = Command.get_cls(line, get_state() == SpiceStates.CONTROL)
            obj = cls(line, ast)

            # don't append if in subcircuit. it will be added in the subcircuit anyway
            if not_state(SpiceStates.SUBCIRCUIT):
                self._obj_lines.append(obj)
            self._logger.debug(os.linesep + repr(obj))
            match obj:
                case If():
                    state_stack.append(SpiceStates.IF)
                    append(obj)
                    raise NotImplementedError
                case Include():
                    self._includes.append(obj.path)
                case Library():
                    self._libs.append(obj)
                case Control():
                    state_stack.append(SpiceStates.CONTROL)
                    control.append(obj)
                case End():
                    # ignore line after
                    break
                case EndControl():
                    if not_state(SpiceStates.CONTROL):
                        raise ParseError(".endc without .control")
                    state_stack.pop()
                case EndIf():
                    if not not_state(SpiceStates.IF):
                        raise ParseError(".endif without .if")
                    state_stack.pop()
                case EndSubcircuit():
                    if not_state(SpiceStates.SUBCIRCUIT):
                        raise ParseError(".ends without .subckt")
                    state_stack.pop()
                    subcircuit = None
                case Library():
                    self._includes.append(obj)
                case Model():
                    self._models.append(obj)
                    append(obj)
                case Subcircuit():
                    state_stack.append(SpiceStates.SUBCIRCUIT)
                    subcircuit = obj
                    self._subcircuits.append(obj)
                    # append(obj)
                case Title():
                    self._titles.append(obj)
                case _:
                    append(obj)
        self._circuit = circuit
        self._control = control

    ##############################################

    def _parse(self) -> None:
        self.parse()
        self.analyse()

    ##############################################

    def parse_string(self, source: str, title_line: bool=True) -> None:
        generator = enumerate(source.splitlines())
        self.read(generator, title_line)
        self._parse()

    ##############################################

    def parse_file(self, path: str | Path, section: str=None) -> None:
        with open(path, 'r', encoding='utf-8') as fh:
            generator = enumerate(fh.readlines())
            self.read(generator, title_line=True, section=section)
            self._parse()

    ##############################################

    def to_spice(self, comment: bool=False, line_length_max: int=None) -> str:
        """Retranslate for each line the AST in Spice syntax."""

        def split_line(line: str, prefix: str) -> str:
            if line_length_max is not None and len(line) > line_length_max:
                position = line.rfind(' ', 0, line_length_max)
                # if position == -1:
                #     position = line_length_max
                # avoid infinite loop
                if position > 0:
                    line = line[:position] + os.linesep + prefix + split_line(line[position:].strip(), prefix)
            return line

        text = ''
        for line, ast in zip(self._lines, self._ast_lines):
            if ast is not None:
                ast_str = str(ast)
                if line.comment:
                    ast_str += f' ; {line.comment}'
                if not (isinstance(ast, Ast.DotCommand) and ast.name in ('.title', '.include', '.lib')):
                    ast_str = split_line(ast_str, '+')
            elif line.comment:
                # comment = line.comment.replace(os.linesep, ' ')
                # ast_str = f'* {comment}'
                # ast_str = split_line(ast_str, '* ')
                comment = line.comment.replace(os.linesep, os.linesep + '* ')
                if line.start == 0:
                    ast_str = comment
                else:
                    ast_str = f'* {comment}'
            text += ast_str + os.linesep
        return text

    ##############################################

    @property
    def title(self) -> SpiceLine | None:
        if self._title_line is not None:
            return self._title_line.comment
        else:
            return None

    @property
    def lines(self) -> Iterator[str]:
        return iter(self._lines)

    @property
    def ast_lines(self) -> Iterator[AstNode]:
        return iter(self._ast_lines)

    @property
    def obj_lines(self) -> Iterator[Command]:
        return iter(self._obj_lines)

    @property
    def titles(self) -> Iterator[Title]:
        return iter(self._titles)

    @property
    def includes(self) -> Iterator[Include]:
        return iter(self._includes)

    @property
    def libraries(self) -> Iterator[Library]:
        return iter(self._libs)

    @property
    def models(self) -> Iterator[Model]:
        return iter(self._models)

    @property
    def subcircuits(self) -> Iterator[Subcircuit]:
        return iter(self._subcircuits)

    @property
    def circuit(self) -> Netlist:
        return iter(self._circuit)

    @property
    def control(self) -> Iterator[Command]:
        return iter(self._control)

    ##############################################

    @property
    def is_only_subcircuit(self) -> bool:
        # Fixme:
        # return not self._circuit and
        return self._subcircuits and not self._models

    ##############################################

    @property
    def is_only_model(self) -> bool:
        # Fixme:
        # return not self._circuit and
        return not self._subcircuits and self._models

####################################################################################################

class SpiceFile(SpiceSource):

    ##############################################

    def __init__(self, path: str | Path, section: str=None) -> None:
        super().__init__()
        self._path = Path(path)
        self.parse_file(path, section)



####################################################################################################

from pathlib import Path
import pickle
# import sha

####################################################################################################

import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from SimulateCircuit import simulate_circuit

from diskcache import Cache

####################################################################################################

analysis = simulate_circuit()

directory = Path('simulation-cache')
cache = Cache(directory)
print(cache.directory)
cache['key'] = analysis
cache.close()

####################################################################################################

# data = pickle.dumps(analysis)

# import codecs
# import os
# hex_name = codecs.encode(os.urandom(16), 'hex').decode('utf-8')
#   '688641106aa7371cf922696cff3e7646'

# with open('simulation-cache/fd/40/d3fe94030bf7d085fc1843946f29.val', 'rb') as fh:
#     data = pickle.load(fh)
#     print(data)

####################################################################################################

# /tmp/diskcache-f_zbph_0/cache.db

# simulation-cache/
# simulation-cache/fd
# simulation-cache/fd/40
# simulation-cache/fd/40/d3fe94030bf7d085fc1843946f29.val
# simulation-cache/cache.db

# PRAGMA foreign_keys=OFF;
# BEGIN TRANSACTION;
# CREATE TABLE Settings ( key TEXT NOT NULL UNIQUE, value);
# INSERT INTO Settings VALUES('count',1);
# INSERT INTO Settings VALUES('size',969750);
# INSERT INTO Settings VALUES('hits',0);
# INSERT INTO Settings VALUES('misses',0);
# INSERT INTO Settings VALUES('statistics',0);
# INSERT INTO Settings VALUES('tag_index',0);
# INSERT INTO Settings VALUES('eviction_policy','least-recently-stored');
# INSERT INTO Settings VALUES('size_limit',1073741824);
# INSERT INTO Settings VALUES('cull_limit',10);
# INSERT INTO Settings VALUES('sqlite_auto_vacuum',1);
# INSERT INTO Settings VALUES('sqlite_cache_size',8192);
# INSERT INTO Settings VALUES('sqlite_journal_mode','wal');
# INSERT INTO Settings VALUES('sqlite_mmap_size',67108864);
# INSERT INTO Settings VALUES('sqlite_synchronous',1);
# INSERT INTO Settings VALUES('disk_min_file_size',32768);
# INSERT INTO Settings VALUES('disk_pickle_protocol',5);
# CREATE TABLE Cache ( rowid INTEGER PRIMARY KEY, key BLOB, raw INTEGER, store_time REAL, expire_time REAL, access_time REAL, access_count INTEGER DEFAULT 0, tag BLOB, size INTEGER DEFAULT 0, mode INTEGER DEFAULT 0, filename TEXT, value BLOB);
# INSERT INTO Cache VALUES(1,'key',1,1637861477.2621181011,NULL,1637861477.2621181011,0,NULL,969750,4,'fd/40/d3fe94030bf7d085fc1843946f29.val',NULL);
# CREATE UNIQUE INDEX Cache_key_raw ON Cache(key, raw);
# CREATE INDEX Cache_expire_time ON Cache (expire_time);
# CREATE INDEX Cache_store_time ON Cache (store_time);
# CREATE TRIGGER Settings_count_insert AFTER INSERT ON Cache FOR EACH ROW BEGIN UPDATE Settings SET value = value + 1 WHERE key = "count"; END;
# CREATE TRIGGER Settings_count_delete AFTER DELETE ON Cache FOR EACH ROW BEGIN UPDATE Settings SET value = value - 1 WHERE key = "count"; END;
# CREATE TRIGGER Settings_size_insert AFTER INSERT ON Cache FOR EACH ROW BEGIN UPDATE Settings SET value = value + NEW.size WHERE key = "size"; END;
# CREATE TRIGGER Settings_size_update AFTER UPDATE ON Cache FOR EACH ROW BEGIN UPDATE Settings SET value = value + NEW.size - OLD.size WHERE key = "size"; END;
# CREATE TRIGGER Settings_size_delete AFTER DELETE ON Cache FOR EACH ROW BEGIN UPDATE Settings SET value = value - OLD.size WHERE key = "size"; END;
# COMMIT;



#! /usr/bin/env python3
# -*- Python -*-

####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

from pathlib import Path
import argparse
import glob
import os
import subprocess
import sys

####################################################################################################

def run_example(file_name, no_gui=True):
    """Run a single example file with proper error handling."""
    print(f"Running {file_name}{' (GUI and plots suppressed)' if no_gui else ''}")
    
    env = os.environ.copy()
    if no_gui:
        env['MPLBACKEND'] = 'Agg'  # Non-interactive matplotlib backend
        env['InSpice_NO_DISPLAY'] = '1'  # Custom env var that can be checked in examples
    env['PYTHONUNBUFFERED'] = '1'  # Ensure output is displayed immediately
    
    try:
        result = subprocess.run(['python', file_name], env=env, check=True)
        return True
    except subprocess.CalledProcessError as e:
        print(f"Error running {file_name}: {e}")
        return False

# Parse command line arguments
parser = argparse.ArgumentParser(description='Run InSpice examples')
parser.add_argument('--no-gui', action='store_true', help='Run without GUI and plots')
args = parser.parse_args()

examples_path = Path(__file__).resolve().parent

successful_examples = []
failed_examples = []

for topic in os.listdir(examples_path):
    topic_path = examples_path.joinpath(topic)
    if not topic_path.is_dir() or topic.startswith('.'):
        continue
        
    python_files = glob.glob(str(topic_path.joinpath('*.py')))
    for file_name in python_files:
        
        # Run automatically if --no-gui is specified, otherwise prompt the user
        if args.no_gui:
            success = run_example(file_name, args.no_gui)
            if success:
                successful_examples.append(file_name)
            else:
                failed_examples.append(file_name)
                print("Example failed, but continuing with next example...")
        else:
            print(f'Would you like to run {file_name}? [y/N/q(uit)]')
            choice = sys.stdin.readline().strip().lower()
            
            if choice == 'q':
                print("Exiting the examples runner.")
                sys.exit(0)
            elif choice == 'y':
                success = run_example(file_name, args.no_gui)
                if success:
                    successful_examples.append(file_name)
                else:
                    failed_examples.append(file_name)
                    print("Example failed, but continuing with next example...")

# After all examples are run - always show the summary, regardless of mode
print("\n" + "="*80)
print(f"EXECUTION SUMMARY:")
print(f"Successfully executed: {len(successful_examples)} examples")

if failed_examples:
    print(f"\nFailed to execute: {len(failed_examples)} examples")
    print("Failed examples:")
    for example in failed_examples:
        print(f"  - {example}")
else:
    print("\nAll examples executed successfully!")
print("="*80)



#skip#

# https://github.com/xesscorp/skidl/blob/master/examples/spice-sim-intro/spice-sim-intro.ipynb

#  N_1 --- A1/adc --- A2/buf --- A3/dac ---
#  |                                       |
#  V1                                      R1
#  |                                       |
#  0                                    ---

# .title Test
# V1 N_1 0 DC 0V AC 1V SIN(1.65V 1.65V 100MegHz 0s 0Hz)
# A1 [N_1] [N_2] adc
# A2 N_2 N_3 buf
# A3 [N_3] [N_4] dac
# R1 N_4 0 1kOhm
# .model adc adc_bridge (fall_delay=1e-09s in_high=0.1V in_low=0.05V rise_delay=1e-09s)
# .model buf d_buffer (fall_delay=1e-09s input_load=1e-12s rise_delay=1e-09s)
# .model dac dac_bridge (out_high=3.3V out_low=1.0V)

# waveforms = sim.transient(step_time=0.1@u_ns, end_time=50@u_ns)

####################################################################################################

# cf. chapter 12

# Ngspice includes a library of predefined ‘Code Models’ that can be placed within any circuit
# description in a manner similar to that used to place standard device models.  Code model instance
# cards always begin with the letter ‘A’, and always make use of a .MODEL card to describe the code
# model desired. Section 28 of this document goes into greater detail as to how a code model similar
# to the predefined models may be developed, but once any model is created and linked into the
# simulator it may be placed using one instance card and one .MODEL card (note here we conform to
# the SPICE custom of referring to a single logical line of information as a ‘card’). As an example,
# the following uses a predefined ‘gain’ code model taking as an input some value on node 1,
# multiplies it by a gain of 5.0, and outputs the new value to node 2. Note that, by convention,
# input ports are specified first on code models. Output ports follow the inputs.

# Example:
# a1 1 2 amp
# .model amp gain(gain=5.0)

# In this example the numerical values picked up from single-ended (i.e. ground referenced) input
# node 1 and output to single-ended output node 2 will be voltages, since in the Interface
# Specification File for this code model (i.e., gain), the default port type is specified as a
# voltage (more on this later). However, if you didn’t know this, the following modifications to the
# instance card could be used to insure it:

# Example:
# a1 %v(1) %v(2) amp
# .model amp gain(gain =5.0)

# The specification %v preceding the input and output node numbers of the instance card indicate to
# the simulator that the inputs to the model should be single-ended voltage values. Other
# possibilities exist, as described later.  Some of the other features of the instance and .MODEL
# cards are worth noting. Of particu- lar interest is the portion of the .MODEL card that specifies
# gain=5.0. This portion of the card assigns a value to a parameter of the ‘gain’ model. There are
# other parameters that can be assigned values for this model, and in general code models will have
# several. In addition to numeric values, code model parameters can take non-numeric values (such as
# TRUE and FALSE), and even vector values. All of these topics will be discussed at length in the
# following pages. In general, however, the instance and .MODEL cards that define a code model will
# follow the abstract form described below. This form illustrates that the number of inputs and
# outputs and the number of parameters that can be specified is rel- atively open-ended and can be
# interpreted in a variety of ways (note that angle-brackets ‘<’ and ‘>’ enclose optional inputs):

# Example:

# AXXXXXXX <%v, %i, %vd, %id, %g, %gd, %h, %hd, or %d>
# + <[> <~><%v, %i, %vd, %id, %g, %gd, %h, %hd, or %d>
# + <NIN1 or +NIN1 -NIN1 or "null">
# + <~>... <NIN2.. <]> >
# + <%v ,%i, %vd , %id, %g, %gd, %h, %hd, %d or %vnam>
# + <[> <~> <%v, %i, %vd, %id, %g, %gd, %h, %hd,
#       or %d> <NOUT1 or +NOUT1 -NOUT1>
# + <~>... <NOUT2.. <]>>
# + MODELNAME

# .MODEL MODELNAME MODELTYPE <( PARAMNAME1 = <[> VAL1 <VAL2... <]>> PARAMNAME2.. > )>

# Square brackets ([ ]) are used to enclose vector input nodes.
# In addition, these brackets are used to delineate vectors of parameters.

# The literal string ‘null’, when included in a node list, is interpreted as no connection at that
# input to the model. ‘Null’ is not allowed as the name of a model’s input or output if the model
# only has one input or one output. Also, ‘null’ should only be used to indicate a missing
# connection for a code model; use on other XSPICE component is not interpreted as a missing
# connection, but will be interpreted as an actual node name.

# The tilde, ‘~’, when prepended to a digital node name, specifies that the logical value of that
# node be inverted prior to being passed to the code model. This allows for simple inversion of
# input and output polarities of a digital model in order to handle logically equivalent cases and
# others that frequently arise in digital system design.

# The following example defines a NAND gate, one input of which is inverted:

# a1 [~1 2] 3 nand1
# .model nand1 d_nand(rise_delay=0.1 fall_delay=0.2)

# The optional symbols %v, %i, %vd, etc. specify the type of port the simulator is to expect for the
# subsequent port or port vector. The meaning of each symbol is given in Table 12.1.

# The symbols described in Table 12.1 may be omitted if the default port type for the model is
# desired. Note that non-default port types for multi-input or multi-output (vector) ports must be
# specified by placing one of the symbols in front of EACH vector port. On the other hand, if all
# ports of a vector port are to be declared as having the same non-default type, then a symbol may
# be specified immediately prior to the opening bracket of the vector. The following examples should
# make this clear:

# Port Type Modifiers
# Modifier Interpretation
# %v       represents a single-ended voltage port - one node name or number is expected for each port.
# %i       represents a single-ended current port - one node name or number is expected for each port.
# %g       represents a single-ended voltage-input, current-output (VCCS) port - one node name or number
#          is expected for each port. This type of port is automatically an input/output.
# %h       represents a single-ended current-input, voltage-output (CCVS) port - one node name or number
#          is expected for each port. This type of port is automatically an input/output.
# %d       represents a digital port - one node name or number is expected for each port.
#          This type of port may be either an input or an output.
# %vnam    represents the name of a voltage source, the current through which is taken as an input.
#          This notation is provided primarily in order to allow models defined using SPICE2G6 syntax
#          to operate properly in XSPICE.
# %vd      represents a differential voltage port - two node names or numbers are expected for each port.
# %id      represents a differential current port - two node names or numbers are expected for each port.
# %gd      represents a differential VCCS port - two node names or numbers are expected for each port.
# %hd      represents a differential CCVS port - two node names or numbers are expected for each port.

# Example 1: - Specifies two differential voltage connections, one to nodes 1 & 2, and one to nodes 3 & 4.
#   %vd [1 2 3 4]

# Example 2: - Specifies two single-ended connections to node 1 and at node 2, and one differential
#              connection to nodes 3 & 4.
#   %v [1 2 %vd 3 4]

# Example 3: - Identical to the previous example...parenthesis are added for additional clarity.
#   %v [1 2 %vd(3 4)]

# Example 4: - Specifies that the node numbers are to be treated in the default fashion for the
#              particular model.  If this model had ‘%v” as a default for this port, then this
#              notation would represent four single-ended voltage connections.
#   [1 2 3 4]

# The parameter names listed on the .MODEL card must be identical to those named in the code model
# itself. The parameters for each predefined code model are described in detail in Sections 12.2
# (analog), 12.3 (Hybrid, A/D) and 12.4 (digital). The steps required in order to specify parameters
# for user-defined models are described in Chapter 28.

# 12.1.2 Examples

# The following is a list of instance card and associated .MODEL card examples showing use of
# predefined models within an XSPICE deck:

# a1 1 2 amp
# .model amp gain(in_offset=0.1 gain=5.0 out_offset=-0.01)

# a2 %i[1 2] 3 sum1
# .model sum1 summer(in_offset=[0.1 -0.2] in_gain=[2.0 1.0] out_gain=5.0 out_offset=-0.01)

# a21 %i[1 %vd(2 5) 7 10] 3 sum2
# .model sum2 summer(out_gain=10.0)

# a5 1 2 limit5
# .model limit5 limit(in_offset=0.1 gain=2.5 out_lower_limit=-5.0 out_upper_limit=5.0 limit_range=0.10 fraction=FALSE)

# a7 2 %id(4 7) xfer_cntl1
# .model xfer_cntl1 pwl(x_array=[-2.0 -1.0 2.0 4.0 5.0] y_array=[-0.2 -0.2 0.1 2.0 10.0] input_domain=0.05 fraction=TRUE)

# a8 3 %gd(6 7) switch3
# .model switch3 aswitch(cntl_off=0.0 cntl_on=5.0 r_off=1e6 r_on=10.0 log=TRUE)



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = [
    'Circuit',
    'Simulator',
    'SpiceLibrary',
    'plot',
]

####################################################################################################

__version__ = '1.6.4.0'
GIT_TAG = 'main'

def show_version():
    print('InSpice Version {}'.format(__version__))

from InSpice.Probe.Plot import plot
from InSpice.Spice.Library import SpiceLibrary
from InSpice.Spice.Netlist import  Circuit, SubCircuit, SubCircuitFactory
from InSpice.Spice.Simulator import Simulator



####################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module setups the logging for InSpice.

"""

####################################################################################################

import yaml
import logging
import logging.config
import os
import sys

####################################################################################################

from InSpice.Config import ConfigInstall

####################################################################################################

def setup_logging(
    application_name='InSpice',
    config_file=ConfigInstall.Logging.default_config_file,
    logging_level=None,
):
    """Setup the logging.

    Logging configuration is set by a YAML file given by *config_file*.

    Alternatively we can set the logging level using the environment variable 'InSpiceLogLevel' or
    using *logging_level*, level can be a integer or a string.  The logging level is set by
    precedence using :

    * `logging_level` parameter if not `None`
    * 'InSpiceLogLevel' environment variable set to: CRITICAL, ERROR, WARNING, INFO, DEBUG
    * else :file:`logging.yml` file settings

    Note: logging level `NOTSET = 0`

    """

    logging_config_file_name = ConfigInstall.Logging.find(config_file)
    logging_config = yaml.load(open(logging_config_file_name, 'r'), Loader=yaml.SafeLoader)

    # YAML fixes

    # Fixme: \033 is not interpreted in YAML
    if ConfigInstall.OS.on_linux:
        formatter_config = logging_config['formatters']['ansi']['format']
        logging_config['formatters']['ansi']['format'] = formatter_config.replace('<ESC>', '\033')

    # Use "simple" formatter for Windows and OSX
    # and "ansi" for Linux
    if ConfigInstall.OS.on_windows or ConfigInstall.OS.on_osx:
        formatter = 'simple'
    else:
        formatter = 'ansi'
    logging_config['handlers']['console']['formatter'] = formatter

    # Load YAML settings
    logging.config.dictConfig(logging_config)

    # Customise logging level
    logger = logging.getLogger(application_name)
    if logging_level is None and 'InSpiceLogLevel' in os.environ:
        level_name = os.environ['InSpiceLogLevel']
        try:
            logging_level = getattr(logging, level_name.upper())
        except AttributeError:
            sys.exit(f'InSpiceLogLevel environment variable is set to an invalid logging level "{level_name}"')
    if logging_level:
        # level can be int or string
        logger.setLevel(logging_level)
    # else use logging.yml

    return logger



###################################################################################################
#
# InSpice - A Spice Package for Python
# Copyright (C) 2021 Fabrice Salvaire
# Copyright (C) 2025 Innovoltive
# Modified by Innovoltive on April 18, 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = ['Simulator']

####################################################################################################

"""This module provides the base class for simulator and a factory method.

"""

####################################################################################################

import logging

####################################################################################################

from ..Config import ConfigInstall
from .Simulation import Simulation

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

# Fixme: DOC: Each analysis mode is performed by a method that return the measured probes.

class Simulator:

    """Base class to implement a simulator.

    """

    _logger = _module_logger.getChild('Simulator')

    #: Define the default simulator
    DEFAULT_SIMULATOR = None
    if ConfigInstall.OS.on_windows:
        DEFAULT_SIMULATOR = 'ngspice-shared'
    else:
        # DEFAULT_SIMULATOR = 'ngspice-subprocess'
        DEFAULT_SIMULATOR = 'ngspice-shared'
        # DEFAULT_SIMULATOR = 'xyce-serial'
        # DEFAULT_SIMULATOR = 'xyce-parallel'

    SIMULATORS = (
        'ngspice',
        'ngspice-shared',
        'ngspice-subprocess',
        'xyce',
        'xyce-serial',
        'xyce-parallel',
    )

    SIMULATOR = None   # for subclass

    ##############################################

    @classmethod
    def factory(cls, *args, **kwargs):

        """Factory to instantiate a simulator.

        By default, it instantiates the simulator defined in :obj:`DEFAULT_SIMULATOR`, however you
        can set the simulator using the :obj:`simulator` parameter.

        Available simulators are:

        * :code:`ngspice` **alias for shared**
        * :code:`ngspice-shared` **DEFAULT**
        * :code:`ngspice-subprocess`
        * :code:`xyce` **alias for serial**
        * :code:`xyce-serial`
        * :code:`xyce-parallel`

        Return a :obj:`InSpice.Spice.Simulator` subclass.

        """

        # Fixme: purpose ??? simplify import...

        simulator = kwargs.pop('simulator', cls.DEFAULT_SIMULATOR)
        sub_cls = None

        if simulator not in cls.SIMULATORS:
            raise NameError(f"Unknown simulator {simulator}")

        if simulator.startswith('ngspice'):
            if simulator == 'ngspice-subprocess':
                from .NgSpice.Simulator import NgSpiceSubprocessSimulator
                sub_cls = NgSpiceSubprocessSimulator
            elif simulator in ('ngspice', 'ngspice-shared'):
                from .NgSpice.Simulator import NgSpiceSharedSimulator
                sub_cls = NgSpiceSharedSimulator

        elif simulator.startswith('xyce'):
            from .Xyce.Simulator import XyceSimulator
            sub_cls = XyceSimulator
            if simulator == 'xyce-parallel':
                kwargs['parallel'] = True

        if sub_cls is not None:
            obj = sub_cls(*args, **kwargs)
            obj._AS_SIMULATOR = simulator
            return obj
        else:
            raise ValueError('Unknown simulator')

    ##############################################

    def __getstate__(self):
        # Pickle: protection for cffi
        return self.__class__.__name__

    ##############################################

    def simulation(self, circuit, **kwargs):
        """Create a new simulation for the circuit.

        Return a :obj:`InSpice.Spice.Simulation` instance`

        """
        # Note: simulation is simulator dependent, thus subclass this method if needed
        return Simulation(self, circuit, **kwargs)

    ##############################################

    @property
    def name(self):
        return self._AS_SIMULATOR

    @property
    def version(self):
        raise NotImplementedError

    ##############################################

    def customise(self, simulation):
        """Customise the simulation"""
        pass

    ##############################################

    def run(self, simulation):
        """Run the simulation and return the waveforms."""
        raise NotImplementedError



from InSpice import Circuit, Simulator
from InSpice.Unit import *
import numpy as np

# Build wheatstone bridge circuit
circuit = Circuit('wheatstone bridge')
circuit.V('input', 'inp', circuit.gnd, 5@u_V)

circuit.R(1, 'inp', 'n1', 5@u_Ω)
circuit.R(2, 'inp', 'n2', 5@u_Ω)
circuit.R(3, 'n1', circuit.gnd, 5@u_Ω)
circuit.R(4, 'n2', circuit.gnd, 5@u_Ω)

# middle resistor
circuit.R(5, 'n1', 'n2', 5@u_Ω)

# adding current probe
circuit.R5.plus.add_current_probe(circuit)
simulator = Simulator.factory()
simulation = simulator.simulation(circuit, temperature=25, nominal_temperature=25)
analysis = simulation.operating_point()

# Print the branches dictionary
print("Values in analysis.branches:")
print(analysis.branches)

# Print the actual current values
print("\nExtracted current values:")
for name, branch in analysis.branches.items():
    current = float(branch)
    print(f"{name}: {current} A")



####################################################################################################

import math
import matplotlib.pyplot as plt

####################################################################################################

import InSpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from InSpice.Doc.ExampleTools import find_libraries
from InSpice import SpiceLibrary, Circuit, Simulator, plot
from InSpice.Unit import *
####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)


####################################################################################################
circuit = Circuit("PMSM Example")
circuit.include(spice_library['pmsm'])  # Include the PMSM library
circuit.include(spice_library['clark'])  # Include the PMSM library
circuit.include(spice_library['iclark'])  # Include the PMSM library
circuit.include(spice_library['park'])  # Include the PMSM load library

#TODO: show how to ramp the input voltage
# circuit.V('qs_ref', 'qs_ref', circuit.gnd, 11.2*math.sqrt(2)@u_V)  # Reference voltage for q-axis
circuit.V('qs_ref', 'qs_ref', circuit.gnd, raw_spice='PWL(0 0 100m {11.2*sqrt(2)})')
circuit.V('ds_ref', 'ds_ref', circuit.gnd, 0@u_V)  # Reference voltage for d-axis
# Xpark   qs_ref ds_ref theta alpha_ref beta_ref park
circuit.X('park', 'park', 'qs_ref', 'ds_ref', 'theta', 'alpha_ref', 'beta_ref')
# Xiclark alpha_ref beta_ref asd bsd csd iclark
circuit.X('iclark', 'iclark', 'alpha_ref', 'beta_ref', 'asd', 'bsd', 'csd')

neutral = 'n'
circuit.NonLinearVoltageSource('as', 'phase_a', neutral, raw_spice='value={v(asd)}')  
circuit.NonLinearVoltageSource('bs', 'phase_b', neutral, raw_spice='value={v(bsd)}')  
circuit.NonLinearVoltageSource('cs', 'phase_c', neutral, raw_spice='value={v(csd)}')  
circuit.R('n', neutral, circuit.gnd, 1@u_GOhm)  # Neutral resistor to ground
circuit.C('n', neutral, circuit.gnd, 1@u_nF)  # Neutral capacitor to ground
# Behavioral voltage source for mechanical angle and speed
# Three-phase voltage sources with proper phase relationships
# #TODO: implement phase in the voltage source. we make cosine to match my course. we add an offset to the delay to match the cosine wave
# vas=circuit.SinusoidalVoltageSource('vas', 'phase_a', circuit.gnd, amplitude=AMPLITUDE, frequency=FLINE, delay=-5*TLINE/4)
# vbs=circuit.SinusoidalVoltageSource('vbs', 'phase_b', circuit.gnd, amplitude=AMPLITUDE, frequency=FLINE, delay=   TLINE/3 - 5*TLINE/4)
# vcs=circuit.SinusoidalVoltageSource('vcs', 'phase_c', circuit.gnd, amplitude=AMPLITUDE, frequency=FLINE, delay= 2*TLINE/3 - 5*TLINE/4)

circuit.V('as', 'phase_a', 'pha', 0) # to measure current in phase and model cable resistance
circuit.V('bs', 'phase_b', 'phb', 0) # to measure current in phase and model cable resistance
circuit.V('cs', 'phase_c', 'phc', 0) # to measure current in phase and model cable resistance

# Add PMSM subcircuit                        
#+rs=3.4 ls=12.1e-3 poles=4 
# +lambda_m=0.0827 Tl=0 J=5e-4 Bm=1e-9
circuit.X('M', 'pmsm', 'pha', 'phb', 'phc', 'theta', 'rpm', 
          rs=0.1,
          ls=1e-3,
          poles=2,
          lambda_m=0.0827,
          J=5e-3,
          Bm=1e-9,
          TL=0.4)
# Add load resistors to complete the circuit and prevent floating nodes


simulator = Simulator.factory()
simulation = simulator.simulation(circuit)
# simulation.options('RSHUNT = 1e12') # helps with convergence in some cases
simulation.options('SAVECURRENTS') # save all the currents in the simulation
# simulation.options('NOINIT')
# simulation.options('KLU')
# simulation.options('ABSTOL=1e-10')  # Absolute tolerance for convergence
# simulation.options('RELTOL=0.01')  # Relative tolerance for convergence
analysis = simulation.transient(step_time=0.1@u_ms, end_time=2@u_s)


##################################################### Plotting #####################################################
figure1, (ax1, ax2, ax3, ax4) = plt.subplots(4, 1, figsize=(10, 8), sharex=True)

ax1.set_title('Three-Phase PMSM Voltage Sources')
ax1.set_ylabel('Voltage [V]')
ax1.grid()
ax1.plot(analysis['phase_a'], label='Phase A')
ax1.plot(analysis['phase_b'], label='Phase B')
ax1.plot(analysis['phase_c'], label='Phase C')
ax1.legend()

ax2.set_title('Three-Phase PMSM Currents')
ax2.set_xlabel('Time [s]')
ax2.set_ylabel('Current [A]')
ax2.grid()
ax2.plot(analysis.branches['vas'], label='Current Phase A')
ax2.plot(analysis.branches['vbs'], label='Current Phase B')
ax2.plot(analysis.branches['vcs'], label='Current Phase C')
ax2.legend()

ax3.set_title('PMSM Mechanical Parameters')
ax3.set_xlabel('Time [s]')
ax3.set_ylabel('Mechanical Parameters')
ax3.grid()
# ax3.plot(analysis['xm.alpha'], label='valpha', color='C0')
# ax3.plot(analysis['xm.beta'], label='vbeta', color='C1')
# ax3.plot(analysis['xm.qs'], label='vqs', color='C0')
# ax3.plot(analysis['xm.ds'], label='vds', color='C1')
# ax3.plot(analysis['xm.wm'], label='Mechanical Angular Velocity [rad/s]', color='C0')
# ax3.plot(analysis['xm.we'], label='Electrical Angular Velocity [rad/s]', color='C1')
ax3.plot(analysis['rpm'], label='Speed [RPM]', color='C0')
# ax3.plot(analysis['xm.theta'], label='Mechanical Angle [rad]', color='C0')
# ax3.plot(analysis['xm.theta2'], label='Electrical Angle [rad]', color='C1')
ax3.legend()

ax4.set_title('PMSM Electrical Parameters')
ax4.set_xlabel('Time [s]')
ax4.set_ylabel('Electrical Parameters')
ax4.grid()
# ax4.plot(analysis['xm.q3'], label='back EMF q-axis [V]', color='C0')
# ax4.plot(analysis['xm.iqs'], label='iqs current [A]', color='C1')
# ax4.plot(analysis['xm.ids'], label='ids current [A]', color='C2')
ax4.plot(analysis['xm.te'], label='Torque [Nm]', color='C0')
ax4.plot(analysis['xm.tl'], label='Load Torque [Nm]', color='C1')
ax4.legend()
plt.tight_layout()
plt.show()

